#ifndef ATOM_H
#define ATOM_H

namespace Atom
{
	bool Init ();
	void Exit ();

	BYTE In (WORD wPort_);
	void Out (WORD wPort_, BYTE bVal_);

// Status register bits
#define STATUS_BUSY		0x80	// Busy - no host access to Command Block Registers
#define STATUS_DRDY		0x40	// Drive is ready to accept commands
#define STATUS_DWF		0x20	// Drive write fault
#define STATUS_DSC		0x10	// Drive seek complete
#define STATUS_DRQ		0x08	// Data request - drive ready to send or receive data
#define STATUS_CORR		0x04	// Correctable data error encountered and corrected
#define STATUS_INDEX	0x02	// Index mark detected on disk (once per revolution)
#define STATUS_ERROR	0x01	// Previous command ended in error

// Error bits in error register
#define ERROR_BBK		0x80	// Bad block mark detected in the requested sector's ID field
#define ERROR_UNC		0x40	// Uncorrectable data error encountered
#define ERROR_UNUSED5	0x20	// Not used
#define ERROR_IDNF		0x10	// Requested sector's ID field not found
#define ERROR_UNUSED3	0x08	// Not used
#define ERROR_ABRT		0x04	// Command aborted due to drive status error or invalid command
#define ERROR_TK0NF		0x02	// Track 0 not found during execution of Recalibrate command
#define ERROR_AMNF		0x01	// Data address mark not found after correct ID field found

/*
#define ERROR_NOERROR	0x01	// No error detected
#define ERROR_FORMATTER	0x02	// Formatter device error
#define ERROR_SECTBUF	0x03	// Sector buffer error
#define ERROR_ECCCIRC	0x04	// ECC circuitry error
#define ERROR_CPU		0x05	// Controller microprocessor error
*/
#define ERROR_DRIVE1	0x80	// Value to or with errors for 2nd drive

// ATA 'identify' command response structure
typedef struct
{
	WORD	wCaps;				// Bit mapped general configuration information. True when bit set
								// Bit 15: Reserved for non magnetic drives
								// Bit 14: Format speed tolerance gap not required
								// Bit 13: Track offset option not available
								// Bit 12: Data strobe offset option not available
								// Bit 11: Rotational speed tolerance is < 0.5%
								// Bit 10: Disk transfer rate not > 10 MB/s
								// Bit 09: Disk transfer rate > 5 MB/s and < 10 MB/s
								// Bit 08: Disk transfer rate > 5 MB/s
								// Bit 07: Removable cartridge drive
								// Bit 06: Fixed drive
								// Bit 05: Spindle motor control option not implemented
								// Bit 04: Head switch time > 15 microseconds
								// Bit 03: Not MFM encoded
								// Bit 02: Not soft sectored
								// Bit 01: Hard Sectored
								// Bit 00: Reserved

	WORD	wLogicalCylinders;	// Number of logical cylinders in the default translation mode

	WORD	wReserved;			// Reserved

	WORD	wLogicalHeads;		// Number of logical heads in the default translation mode
	WORD	wBytesPerTrack;		// Number of unformatted bytes per logical track
	WORD	wBytesPerSector;	// Number of unformatted bytes per sector
	WORD	wSectorsPerTrack;	// Number of logical sectors per track

	WORD	wInterSectorGaps;	// Bits 15...08: Inter Sector Gap after Index & before splice
								// Bits 07...00: Inter Sector Gap bytes
	WORD	wPhaseLockOscil;	// Bits 15...08: Reserved
								// Bits 07...00: Bytes in Phase Lock Oscillator field

	WORD	wVendorStatusWords;	// Number of vendor unique status words

	char	szSerialNumber[20];	// Serial number, 20 ASCII chars, right aligned & padded with 20h

	WORD	wControllerType;	// 0000h: Not specified
								// 0001h: Single ported, single sector buffer capable of data transfers only
								//			to or from the host or the disk at one time
								// 0002h: Dual ported, multiple sector buffer capable of simultaneous data
								//			transfers to and from the host, or from the host and the disk
								// 0003h: Dual ported, multiple sector buffer capable of simultaneous data
								//			transfers with read caching
								// 0004h-FFFFh: Reserved

	WORD	wBufferSize512;		// Buffer size in 512 byte increments

	WORD	wLongECCBytes;		// Number of ECC bytes passed to host on R/W long operations

	char	szFirmwareRev[8];	// Firmware revision, 8 ASCII chars, left aligned & space padded

	char	szModelNumber[40];	// Model Number, 40 ASCII chars, left aligned & space padded

	WORD	wReadWriteMulti;	// READ/WRITE multiples implemented

	WORD	wDwordIO;			// Supports double word I/O transfer

	WORD	wReserved2;			// Reserved
	WORD	wReserved3;			// Reserved

	WORD	wPIOTransferTime;	// Minimum PIO data transfer cycle time in nsec
	WORD	wDMATransferTime;	// Minimum DMA data transfer cycle time in nsec
}
DRIVEIDENTITY;

};	// namespace Atom

#endif
