#ifndef CDRIVE_H
#define CDRIVE_H

#include "VL1772.h"
#include "CDisk.h"

class CDrive
{
	public:
		CDrive () { ResetAll(); }
		virtual ~CDrive () { if (IsInserted()) Eject(); }

	public:
		bool IsInserted () const { return m_pDisk != NULL; }
		bool IsModified () const { return m_pDisk != NULL && m_pDisk->IsModified(); }
		bool IsMotorOn () const { return (m_sRegs.bStatus & MOTOR_ON) != 0; }
		bool IsLightOn () const { return IsMotorOn() && m_pDisk; }

		bool Insert (const char* pcszSource_);
		void Insert (CDisk* pDisk_);
		CDisk* Eject (bool fDelete=true);
		bool Save ();

		// Disk I/O calls used to drive us
		BYTE In (WORD wPort_);
		void Out (WORD wPort_, BYTE bVal_);
		void FrameEnd ();

		// Functions to deal with the contents of the disk
		bool FindSector  (int nSide_, int nTrack_, int nSector_, IDFIELD* pIdField_=NULL);
		BYTE ReadAddress (int nSide_, int nTrack_, IDFIELD* pIdField_);
		UINT ReadTrack (int nSide_, int nTrack_, BYTE* pbTrack_, UINT uSize_);
		BYTE VerifyTrack (int nSide_, int nTrack_);

	protected:
		CDisk*		m_pDisk;		// The disk currently inserted in the drive, if any
		VL1772Regs	m_sRegs;		// VL1772 controller registers
		int			m_nHeadPos;		// Physical track the drive head is above

		BYTE		m_abBuffer[MAX_TRACK_SIZE];	// Buffer big enough for anything we'll read or write
		BYTE*		m_pbBuffer;
		UINT		m_uBuffer;
		BYTE		m_bDataStatus;	// Status value for end of data, where the data CRC can be checked

		DWORD		m_dwLastMotorOn;

	protected:
		void ResetAll();
		void ModifyStatus (BYTE bEnable_, BYTE bReset_);
		void SetMotor (bool fOn_);
};

extern CDrive g_Drive1, g_Drive2;

#endif
