#ifndef CSTREAM_H
#define CSTREAM_H

#ifdef USE_ZLIB
#define ZLIB_DLL
#include "zlib.h"		// for gzopen, gzclose, etc.
#include "unzip.h"		// for unzOpen, unzClose, etc.
#endif


class CStream
{
	public:
		CStream (const char* pcszStream_, bool fReadOnly_=false);
		virtual ~CStream ();

	public:
		static CStream* Open (const char* pcszStream_, bool fReadOnly_=false);

	public:
		bool IsReadOnly () const { return this && m_fReadOnly; }
		const char* GetName () { return m_pcszStream; }

		virtual bool Rewind () = 0;
		virtual LONG Read (void* pvBuffer_, LONG lLen_) = 0;
		virtual LONG Write (void* pvBuffer_, LONG lLen_) = 0;
		
	protected:
		virtual void Close () = 0;

////////////////////////////////////////////////////////////////////////////////

	protected:
		enum { modeClosed, modeReading, modeWriting };
		int			m_nMode;

		bool		m_fReadOnly;
		const char*	m_pcszStream;
};

class CFileStream : public CStream
{
	public:
		CFileStream (FILE* hFile_, const char* pcszStream_, bool fReadOnly_)
			: CStream(pcszStream_, fReadOnly_), m_hFile(hFile_) { }
		virtual ~CFileStream () { Close(); }

	public:
		bool Rewind ();
		LONG Read (void* pvBuffer_, LONG lLen_);
		LONG Write (void* pvBuffer_, LONG lLen_);

	protected:
		FILE* m_hFile;

	protected:
		void Close ();
};


#ifdef USE_ZLIB

class CZLibStream : public CStream
{
	public:
		CZLibStream (gzFile hFile_, const char* pcszStream_, bool fReadOnly_)
			: CStream(pcszStream_, fReadOnly_), m_hFile(hFile_) { }
		virtual ~CZLibStream () { Close(); }

	public:
		bool Rewind ();
		LONG Read (void* pvBuffer_, LONG lLen_);
		LONG Write (void* pvBuffer_, LONG lLen_);

	protected:
		gzFile	m_hFile;

	protected:
		void Close ();
};

class CZipStream : public CStream
{
	public:
		CZipStream (unzFile hFile_, const char* pcszFile_, bool fReadOnly_)
			: CStream(pcszFile_, fReadOnly_), m_hFile(hFile_) { }
		virtual ~CZipStream () { Close(); }

	public:
		bool Rewind ();
		LONG Read (void* pvBuffer_, LONG lLen_);
		LONG Write (void* pvBuffer_, LONG lLen_);

	protected:
		unzFile	m_hFile;

	protected:
		void Close ();
};

#endif	// USE_ZLIB

#endif	// CSTREAM_H
