#ifndef Z80_H
#define Z80_H

#include "SAM.h"

namespace CPU
{
	bool Init (bool InitMemory_=true);
	void Exit ();

	void Run ();
	void UpdateContention ();
};

extern int InterruptType;
extern DWORD g_dwCycleCounter;
extern int g_nLine, g_nLineCycle;


#define OP_NOP						0x00		// Z80 opcode for NOP
#define OP_HALT						0x76		// Z80 opcode for HALT
#define OP_EI						0xfb		// Z80 opcode for EI
#define	OP_DI						0xf3		// Z80 opcode for DI

#define IM1_INTERRUPT_HANDLER		0x0038		// Interrupt mode 1 handler address
#define NMI_INTERRUPT_HANDLER		0x0066		// Non-maskable interrupt handler address

// These have been accurately measured using various tests (ask me if you want more details on that),
// and only these 4 exact values will match what the real SAM gives!
#define INT_ACTIVE_TIME				128			// tstates interrupt is active and will be triggered
#define INT_VISIBLE_TIME			120			// tstates interrupt is visible on the SAM status port
#define IM1_INTERRUPT_TIME			16			// tstates for IM0/1 interrupt to reach handler
#define IM2_INTERRUPT_TIME			24			// tstates for IM2 interrupt to reach handler

// Round a tstate value up to a given power of 2; and so the line total rounds up to the next whole multiple
#define ROUND(t,n)	(((t) + (n)-1) & ~((n)-1))
#define A_ROUND(t,n)	(ROUND(g_nLineCycle+(t),n) - g_nLineCycle)

// Z80 pseudo states to keep track of what we're doing
enum { Z80_none, Z80_nmi, Z80_reset, Z80_pause };

// Bit values for the F register
#define F_CARRY		0x01
#define F_NADD		0x02	// For BCD (DAA)
#define F_PARITY	0x04
#define F_OVERFLOW	0x04
#define F_HCARRY	0x10	// For BCD (DAA)
#define F_ZERO		0x40
#define F_NEG		0x80


// NOTE: ENDIAN-SENSITIVE!
typedef struct
{
	union
	{
		WORD	W;

#ifndef __BIG_ENDIAN
		struct { BYTE l_, h_; } B;	// Little endian
#else
		struct { BYTE h_, l_; } B;	// Big endian (untested!)
#endif
	};
}
REGPAIR;

typedef struct
{
	REGPAIR	AF, BC, DE, HL;
	REGPAIR	AF_, BC_, DE_, HL_;
	REGPAIR	IX, IY;
  	REGPAIR	SP, PC;

	BYTE	I, R;
	BYTE	IFF1, IFF2, IM;
}
Z80Regs;


#endif	// Z80_H
