//	Notes:
//		This is the Z80 to C conversion of the disassembler I wrote for TurboMON.
//		Looking back at it now it seems quite a warped way of doing it!
//
//		In debugging the MIDI OUT implementation I noticed that some opcodes aren't
//		decoded properly, but I've yet to bother looking into why - it all worked
//		fine in TurboMON and it was thoroughly tested back then!
//
//		Once a proper debugger has been adding this will come in quite handy...

#include "stdafx.h"

#include "Util.h"
#include "Memory.h"

// Bit table indicating which opcodes can have a DD/FD index prefix
BYTE abIndexOps[] =
{
	0x08, 0x8A, 0x0A, 0x8A, 0x3E, 0xBE, 0x3E, 0x08, 0x08, 0x8B, 0x0A, 0x4A, 0x3E, 0x3E, 0x3E, 0x08,
	0x08, 0x08, 0x08, 0x08, 0x3E, 0x3E, 0x36, 0x08, 0x00, 0x87/*0x83*/, 0x00, 0x00, 0x3C, 0x3C, 0x3C, 0x00
};

char* string_tab[] =
{
	"\x9f[ADD|ADC|SUB|SBC|AND|XOR|OR|CP] [A,|A,||A,]",  "[IN|DE]C ", "[|!!%c,]",
	"[BC|DE|q|SP]", "\xb3[\x9f[RLC|RRC|RL* |RR* |SLA|SRA|SLL|SRL]|BIT|RES|SET]",
	"\x9fg", "[NZ|Z|NC|C|PO|PE|P|M]", "%i[H\0]q*h", "(q%i[)\0]%d)", "LD ",
	"\xa3[BC|DE|q|AF]", "%i[L\0]q*l", "", "%n", "\xa1([BC|DE])", "(%a)", "%i[HL|IX%j|IY%k]",
	"[B|C|D|E|H|L|i|A]", "[B|C|D|E|h|l|i|A]"
};

char* normal_string = 
	"\xb3[\x87[\xa9[\xa1[\x99[NOP|EX AF,AF']|\x99[DJNZ|JR] %e\2]|JR \x9bg,%e\2]|"
	"\x99[j\xa3" "d,%a\3|ADD q,\xa3" "d]|j\xa9[\x99[o,A|A,o]|\x9b[p,q|q,p|p,A|A,p]\3]|"
	"\x99" "b\xa3" "d|!\x81" "b\x9fs%h[|\5]|j%h[\x9fs,n\2]i,%m\6|"
	"\xa9[R\x99[L|R]\xa1[C]A|\x9b[DAA|CPL|SCF|CCF]]]\1|%h[j%l[\x9fs|"
	"\x9fr],%l[\x87s\1|\x87r]|%l[j\x9fi,\x87r|HALT\1]]\5|a\x87s%l[\1|\5]|"
	"\x87[RET f|\x99[POP k|\xa3[RET|EXX|JP (q)|jSP,q]]|JP f,%a\3|"
	"\x9f[JP %a\3||OUT (n),A\2|IN A,(n)\2|EX (SP),q|EX DE,HL|DI|EI]|CALL f,%a\3|"
	"\x99[PUSH k|CALL %a\3]|a\xa9[n\2]\x9b[|||n\2]%b\2|RST %f]\1]";

char* unused_string = "*[q* PREFIX*]";

char* ed_string =
	"\xb3[|\x87[!\x81[IN|OUT] [%h[\x9fr|X],](C)[|,%h[\x9fr|0]]\2|"
	"\x99[SB|AD]C q,\xa3" "d|j\x99[p,]\xa3" "d\x99[|,p]\4|NEG|RET\x99[N|I]|IM \x9b[0|0**|1|2]|"
	"\xa9[j\xa1[|A,]\x99[I|R]\xa1[,A]|\xa1[R\x99[R|L]D|NOP]]]\2|"
	"\xa9[|\x91[\x83[LD|CP|IN|\xa1[OUT\x99[I|D]\2|OT]]\x99[I|D]\xa1[|R]\2]]]NOP\2";

char* cb_string = "%l[%i[e c\x87r1]\xb3[|e %c,i\3]j\x87r,e* ci\3|e c\x87r%i[1]]\3";


static WORD wPC = 0;
static char szOutput[64], *pszOutput;
static BYTE abOpcode[4], *pbOpcode, bOpcode;
static BYTE abStack[10] = {0,0,0,0,0,0,0,0,0,0}, *pbStack;

int nType = 0;


void skip_section (char** ppszTable)
{
	while (1)
	{
		switch (*(*ppszTable)++)
		{
			case ']':	return;
			case '[':	skip_section(ppszTable);
		}
	}
}

void str_function (BYTE b_)
{
	int i;
	bool fPositive = !(pbOpcode[1] & 0x80);
	BYTE bDisp = fPositive ? pbOpcode[1] : (0 - pbOpcode[1]);

	switch (b_)
	{
		case 'a':	pszOutput += sprintf(pszOutput, "#%02hX%02hX", pbOpcode[2], pbOpcode[1]);	break;
		case 'b':	*pszOutput++ = '%'; for (i = 0 ; i < 8 ; i++) *pszOutput++ = '0' + ((pbOpcode[1] >> i) & 1); break;
		case 'c':	*pszOutput = '0' + ((pbOpcode[0] >> 3) & 7); break;
		case 'd':	if (pbOpcode[1]) pszOutput += sprintf(pszOutput, "%c#%02hX", fPositive ? '+' : '-', bDisp); break;
		case 'e':	pszOutput += sprintf(pszOutput, "#%04hX", wPC + 2 + static_cast<signed char>(pbOpcode[1])); break;
		case 'f':	pszOutput += sprintf(pszOutput, "%hX", pbOpcode[0] & 0x38); break;
		case 'h':	*pbStack = ((pbOpcode[0] >> 3) & 7) == 6; break;
		case 'i':	*pbStack = nType; break;
		case 'l':	*pbStack = (pbOpcode[0] & 7) == 6; break;
		case 'm':	pszOutput += sprintf(pszOutput, "#%02hX", pbOpcode[1 + (!nType ? 0 : 1)]); break;
		case 'n':	pszOutput += sprintf(pszOutput, "#%02hX", pbOpcode[1]); break;
	}
}

int parse_str (char* pszTable)
{
	while (1)
	{
		BYTE b = *pszTable++;

		switch (b)
		{
			case '[':
			{
				int nPos = *pbStack++;

				while (nPos)
				{
					switch (*pszTable++)
					{
						case '[':	skip_section(&pszTable);	break;
						case ']':	nPos = 0;					break;
						case '!':
						case '|':	nPos--;						break;
					}
				}
				pbStack--;

				break;
			}

			case ']':	pbStack--;						continue;
			case '!':	continue;
			case '|':	skip_section(&pszTable);		break;
			case '*':	*pszOutput++ = *pszTable++;		break;
			case '%':	str_function(*pszTable++);		break;
			case ' ':	pszOutput = szOutput+6;			break;

			default:
				if (b >= 0x80)
					*pbStack = (pbOpcode[0] >> ((b >> 3) & 7)) & (b & 7);
				else if (b >= 'a')
					parse_str(string_tab[b-'a']);
				else if (b < 5)
					return nType ? b + 1 : b;
				else if (b <= 6)
					return nType ? b - 2 : b - 4;
				else
					*pszOutput++ = b;
		}
	}

	return 0;
}

int Disassemble (WORD wPC_, char* psz_, int cbSize_)
{
	// Initialise our working variables
	memset(szOutput, ' ', sizeof szOutput);
	pszOutput = szOutput;
	pbOpcode = abOpcode;
	*(pbStack = abStack) = 0;
	wPC = wPC_;

	// Read 4 bytes (longest possibly instruction length) from the current PC location
	for (int i = 0 ; i < sizeof(abOpcode) ; i++)
		abOpcode[i] = read_byte(wPC+i);

	// Check for an IX/IY prefix
	switch (*pbOpcode)
	{
		case 0xdd:	nType = 1;	pbOpcode++;		break;
		case 0xfd:	nType = 2;	pbOpcode++;		break;
		default:	nType = 0;					break;
	}

	// Check for the a prefix for the two extended sets
	char* pszTable = normal_string;
	switch (bOpcode = *pbOpcode)
	{
		case 0xed:	pszTable = ed_string;	pbOpcode++;		break;
		case 0xcb:	pszTable = cb_string;	pbOpcode++;		break;
	}

	// If we have an index prefix, make sure the opcode following it is affected by it
	if (nType && !(abIndexOps[bOpcode & 31] & (1 << ((bOpcode >> 5) & 7))))
		pszTable = unused_string;

	// Parse the instruction, terminate and copy the output to the supplied buffer
	int nLength = parse_str(pszTable);
	*pszOutput = '\0';
	strncpy(psz_, szOutput, min(cbSize_, pszOutput-szOutput+1));

	// Return the actual length of the instruction encountered
	return nLength;
}
