#ifndef IO_H
#define IO_H

#include "SAM.h"

namespace IO
{
	bool Init (bool fReset_=false);
	void Exit (bool fReinit_=false);

	BYTE In (WORD wPort_);
	void Out (WORD wPort_, BYTE bVal_);
};

#define in_byte		IO::In
#define out_byte	IO::Out


#define LEPR_PORT			128
#define HEPR_PORT			129

#define CLOCK_PORT			239			// Input and output for SAMBUS and DALLAS clocks

#define LPEN_PORT			248			// Input
#define HPEN_PORT			504			// Input
#define CLUT_BASE_PORT		248			// Output

#define STATUS_PORT			249			// Input
#define LINE_PORT			249			// Output

#define LMPR_PORT			250			// Input and output
#define HMPR_PORT			251			// Input and output
#define VMPR_PORT			252			// Input and output
#define MIDI_PORT			253			// Input and output

#define KEYBOARD_PORT		254			// Input
#define BORDER_PORT			254			// Output

#define SOUND_MASK			511
#define SOUND_DATA			255			// Register output
#define SOUND_ADDR			511			// Data output
#define ATTR_PORT			255			// Input

// Floppy drives or ATOM hard disk - 111d0srr : d = drive, s = side, r = register
#define FLOPPY_MASK			0xf8		// 11111000
#define FLOPPY1_BASE		224			// 224 to 231
#define FLOPPY2_BASE		240			// 240 to 247

#define PRINTL_MASK			0xfc		// 11111100
#define PRINTL_BASE			232			// 11101000
#define PRINTL1_DATA		232
#define PRINTL1_STAT		233
#define PRINTL2_DATA		234
#define PRINTL2_STAT		235

#define SERIAL_MASK			0xfe		// 11111110
#define SERIAL_BASE			236			// 11101100
#define SERIAL1				236			
#define SERIAL2				237
// Are 238 and 239 SERIAL3 and SERIAL4?  what about the clock port?

#define LMPR_PAGE			(lmpr & 0x1f)
#define LMPR_ROM0_OFF		0x20
#define LMPR_ROM1			0x40
#define LMPR_WPROT			0x80
#define HMPR_PAGE			(hmpr & 0x1f)
#define HMPR_MD3COL			((hmpr & 0x60) >> 3)
#define HMPR_MCNTRL			(hmpr & 0x80)

#define VMPR_MODE_MASK		0x60
#define VMPR_SCREEN_PAGE	(vmpr & 0x1f)

#define MODE_1				0x00
#define MODE_2				0x20
#define MODE_3				0x40
#define MODE_4				0x60

#define BORD_VAL(x)			((((x) & 0x20 ) >> 2) | ((x) & 0x07))
#define BORD_COL(x)			((x) & 0x27)
#define BORD_SOFF			(border & 0x80)

#define MD3_SWITCH(x)		(((x) == 0x01 || (x) == 0x02) ? (x) ^ 0x03 : (x))

#define LPEN_TXFMST			0x02	// Bit set if MIDI OUT is currently transmitting a byte


// Bits in the status register to RESET to signal the various interrupts
#define STATUS_INT_LINE		0x01
#define STATUS_INT_MOUSE	0x02		// Not currently used in SimCoupe as everything seems to poll the mouse
#define STATUS_INT_MIDIIN	0x04
#define STATUS_INT_FRAME	0x08
#define STATUS_INT_MIDIOUT	0x10
#define STATUS_INT_NONE		0xff

// MIDI transfer rates are 31.25 Kbaud. Data has 1 start bit, 8 data bits, and 1 stop bit, for 320us serial byte.
#define MIDI_TRANSMIT_TIME		USECONDS_TO_TSTATES(320)
#define MIDI_OVERLAP_TIME		USECONDS_TO_TSTATES(16)
#define MIDI_INT_ACTIVE_TIME	(INT_ACTIVE_TIME >> 1)


// Helper macros for SAM keyboard matrix manipulation
#define IsSamKeyPressed(n)		(!(keyports[n >> 3] & (1 << (n & 7))))
#define PressSamKey(n)			(keyports[n >> 3] &= ~(1 << (n & 7)))
#define ReleaseSamKey(n)		(keyports[n >> 3] |= (1 << (n & 7)))
#define ToggleSamKey(n)			(keyports[n >> 3] ^= (1 << (n & 7)))
#define ReleaseAllSamKeys()		memset(keyports, 0xff, sizeof keyports);

// Key constants used with the key macros above
enum eSamKey
{
	SK_SHIFT, SK_Z, SK_X, SK_C, SK_V, SK_F1, SK_F2, SK_F3,
	SK_A, SK_S, SK_D, SK_F, SK_G, SK_F4, SK_F5, SK_F6,
	SK_Q, SK_W, SK_E, SK_R, SK_T, SK_F7, SK_F8, SK_F9,
	SK_1, SK_2, SK_3, SK_4, SK_5, SK_ESCAPE, SK_TAB, SK_CAPS,
	SK_0, SK_9, SK_8, SK_7, SK_6, SK_MINUS, SK_PLUS, SK_DELETE,
	SK_P, SK_O, SK_I, SK_U, SK_Y, SK_EQUALS, SK_QUOTES, SK_F0,
	SK_RETURN, SK_L, SK_K, SK_J, SK_H, SK_SEMICOLON, SK_COLON, SK_EDIT,
	SK_SPACE, SK_SYMBOL, SK_M, SK_N, SK_B, SK_COMMA, SK_PERIOD, SK_INV,
	SK_CONTROL, SK_UP, SK_DOWN, SK_LEFT, SK_RIGHT, SK_NONE, SK_MAX=SK_NONE
};




// The SAM ports

// Read / Write ports

// Paging ports for internal and external memory
extern BYTE vmpr, hmpr, lmpr, lepr, hepr;
extern BYTE vmpr_mode, vmpr_page1, vmpr_page2;

extern BYTE mode3_bcd48;

extern BYTE border;
extern BYTE border_col;

// Write only ports
extern BYTE line_int;
extern UINT clut[N_CLUT_REGS], clutval[N_CLUT_REGS];

// Read only ports
extern BYTE status_reg, status_reg_active;
extern BYTE lpen, hpen;


#endif
