#include "stdafx.h"
#include "Main.h"

#include "CPU.h"
#include "Memory.h"
#include "Options.h"
#include "Util.h"
#include "Sound.h"
#include "Input.h"
#include "Video.h"
#include "Frame.h"
#include "UI.h"

#include "CDrive.h"
extern CDrive g_Drive1, g_Drive2;


// Use default main unless the platform requires something special
#ifndef CUSTOM_MAIN
#include "Disassem.h"
int main (int argc_, char* argv_[])
{
	if (Main::Init(argc_, argv_))
	{
/*
		Temporary disassembler test code!
		for (int b = 0 ; b <= 255 ; b++)
		{
			write_byte(0x4000, 0xcb);
			write_byte(0x4001, b);

			char sz[256];
			int nLen = Disassemble(0x4000, sz, sizeof sz);
			TRACE("%02x %03d  %d  %s\n", b, b, nLen, sz);
		}

		return 0;
*/
		CPU::Run();
		Main::Exit();
	}

	return 0;
}
#endif


namespace Main
{

bool Init (int argc_, char* argv_[])
{
	// Load settings and check command-line options
	if (!Options::Load(argc_, argv_))
		return 0;

	open_log();

	// Initialise all modules, return true if all were successful
	return (UI::Init() && /*Input::Init() &&*/ Sound::Init() && Video::Init() && Frame::Init() && CPU::Init());
}

void Exit ()
{
	Options::Save();

	Input::Exit();
	Frame::Exit();
	IO::Exit();
	Sound::Exit();
	CPU::Exit();
	UI::Exit();
	Video::Exit();

	close_log();
}

};	// namespace Main
