//	Notes:
//		Re-written from original version.  Still has conditional Win32 implementation
//		that will be removed when a generic implementation is written.  Currently
//		non-Win32 platforms use the default values for each option.
//
//		Command-line parsing needs putting back in, to override the options taken
//		from the file.  Might be nice to be able to specify the options filename...

#include "stdafx.h"

#include "Options.h"
#include "Util.h"


namespace Options
{

OPTIONS g_sOptions;

enum { OT_BOOL, OT_INT, OT_STRING };

typedef struct tag
{
	const char*	pcszName;										// Option name used in config file
	int			nType;											// Option type

	union {	void* pv;  char* ppsz;  bool* pf;  int* pn; };		// Address of config variable
	union { DWORD dw;  bool f;  int n;  const char* pcsz; };	// Default value of option
}
OPTION;

#define OPT(n,t,v,d)		{ n, t, (VOID*)&g_sOptions.v, (DWORD)(d) }


OPTION asOptions[] = 
{
	OPT("LogFile",		OT_STRING,	logfile,		""),		// No logfile

	OPT("FastReset",	OT_BOOL,	fastreset,		true),		// Allow fast Z80 resets
	OPT("AutoBoot",		OT_BOOL,	autoboot,		false),		// Don't auto-boot inserted disk
	OPT("Spectrum",		OT_BOOL,	spectrum,		false),		// Not in Spectrum mode

	OPT("Video",		OT_BOOL,	video,			true),		// Video enabled
	OPT("Sync",			OT_BOOL,	sync,			true),		// Sync to 50Hz
	OPT("FrameSkip",	OT_INT,		frameskip,		0),			// Auto frame-skipping
	OPT("SourceSize",	OT_INT,		sourcesize,		1),			// Source display is 1x1
	OPT("Scale",		OT_INT,		scale,			2),			// Windowed display is 2x2
	OPT("Ratio5_4",		OT_BOOL,	ratio5_4,		false),		// Don't use 5:4 screen ratio
	OPT("Fullscreen",	OT_BOOL,	fullscreen,		false),		// Not full screen
	OPT("Depth",		OT_INT,		depth,			8),			// Full screen mode uses 8-bit colour
	OPT("FloppyLeds",	OT_BOOL,	floppyleds,		true),		// Show floppy drive LEDs
	OPT("VideoMethod",	OT_INT,		videomethod,	5),			// Try for the overlay surface, also used as a back-buffer
	OPT("DrawMethod",	OT_INT,		drawmethod,		0),			// Draw in native pixel format instead of the SAM palette test

	OPT("ROM0",			OT_STRING,	rom0,			"sam_rom0.rom"),
	OPT("ROM1",			OT_STRING,	rom1,			"sam_rom1.rom"),
	OPT("SpectrumROM",	OT_STRING,	spectrumrom,	"spectrum.rom"),
	OPT("MainMemory",	OT_INT,		mainmem,		512),		// 512K main memory
	OPT("ExternalMem",	OT_INT,		externalmem,	0),			// No external memory

	OPT("Disk1",		OT_STRING,	disk1,			"disk1"),	// Disk for floppy drive 1
	OPT("Disk2",		OT_STRING,	disk2,			"disk2"),	// Disk for floppy drive 2
	OPT("Drive1",		OT_INT,		drive1,			1),			// Floppy drive 1 present
	OPT("Drive2",		OT_INT,		drive2,			1),			// Floppy drive 2 present

	OPT("KeyboardMode",	OT_INT,		keyboard,		1),			// SAM keyboard mode
	OPT("AltForCntrl",	OT_BOOL,	altforcntrl,	false),		// Left-Alt not used for SAM Cntrl
	OPT("Mouse",		OT_BOOL,	mouse,			false),		// Mouse not present
	OPT("Joystick1",	OT_INT,		joystick1,		1),			// Joystick 1 device number
	OPT("Joystick2",	OT_INT,		joystick2,		2),			// Joystick 2 device number
	OPT("JoystickTol",	OT_INT,		joysticktol,	20),		// Joystick deadzone is 20% around central position

	OPT("Parallel1",	OT_INT,		parallel1,		0),			// Nothing on parallel port 1
	OPT("Parallel2",	OT_INT,		parallel2,		0),			// Nothing on parallel port 2

	OPT("SambusClock",	OT_BOOL,	sambus_clock,	true),		// SAMBUS clock present
	OPT("DallasClock",	OT_BOOL,	dallas_clock,	false),		// DALLAS clock not present

	OPT("Sound",		OT_BOOL,	sound,			true),		// Sound enabled
	OPT("Frequency",	OT_INT,		rate,			22050),		// 22KHz
	OPT("Bits",			OT_INT,		bits,			16),		// 16-bit
	OPT("Stereo",		OT_BOOL,	stereo,			true),		// Stereo
	OPT("Filter",		OT_BOOL,	filter,			false),		// Sound filter disabled (not implemented by SAASOUND yet)
	OPT("Latency",		OT_INT,		latency,		6),			// Sound latency of five frames
	OPT("DacOptimise",	OT_BOOL,	dacoptimise,	false),		// Optimise sound for SAA output

	OPT("Profile",		OT_BOOL,	profile,		true),		// Profile stats

	{ NULL,				0,			0,				0 }
};

inline bool IsTrue (const char* pcsz_)
{
	return pcsz_ && (!strcasecmp(pcsz_, "true") || !strcasecmp(pcsz_, "on") || !strcasecmp(pcsz_, "enabled") || 
					!strcasecmp(pcsz_, "yes") || !strcasecmp(pcsz_, "1"));
}


bool Load (int argc_, char* argv_[])
{
	// Read settings from the INI file, using the default if
	for (OPTION* p = asOptions ; p->pcszName ; p++)
	{
		char szOption[256];
#ifdef _WIN32
		bool fDefault = !GetPrivateProfileString("Options", p->pcszName, "", szOption, sizeof szOption, ".\\SimCoupe.ini");
#else
		bool fDefault = true;
		strncpy(szOption, p->pcszName, sizeof szOption);
#endif

		switch (p->nType)
		{
			case OT_BOOL:		*p->pf = fDefault ? p->f : IsTrue(szOption);	break;
			case OT_INT:		*p->pn = fDefault ? p->n : atoi(szOption);		break;
			case OT_STRING:		strcpy(p->ppsz, fDefault ? p->pcsz : szOption);	break;
		}
	}


	while (--argc_)
	{
		const char* pcszOption = (*++argv_)+1;
		for (OPTION* p = asOptions ; p->pcszName && strcasecmp(p->pcszName, pcszOption); p++);

		if (!p->pcszName || !--argc_)
		{
			TRACE("Invalid option: %s\n", pcszOption);
			continue;
		}

		switch (p->nType)
		{
			case OT_BOOL:		*p->pf = IsTrue(*++argv_);	break;
			case OT_INT:		*p->pn = atoi(*++argv_);	break;
			case OT_STRING:		strcpy(p->ppsz, *++argv_);	break;
		}
	}

	return true;
}


void Save ()
{
#ifdef _WIN32

	for (OPTION* p = asOptions ; p->pcszName ; p++)
	{
		char szInt[10];

		switch (p->nType)
		{
			case OT_BOOL:		WritePrivateProfileString("Options", p->pcszName, *p->pf ? "Yes" : "No", ".\\SimCoupe.ini");	break;
			case OT_INT:		itoa(*p->pn, szInt, sizeof szInt); WritePrivateProfileString("Options", p->pcszName, szInt, ".\\SimCoupe.ini");	break;
			case OT_STRING:		WritePrivateProfileString("Options", p->pcszName, p->ppsz, ".\\SimCoupe.ini");				break;
		}
	}
#endif
}

};	// namespace Options
