#include "stdafx.h"

#include "Input.h"
#include "IO.h"
#include "Util.h"
#include "Options.h"
#include "Mouse.h"
#include "UI.h"

extern bool g_fActive;
extern BYTE keyports[9];

namespace Input
{
#define MAX_JOYSTICKS				8
#define KEYBOARD_BUFFER_SIZE		10

typedef struct
{
	int		nChar;				// Symbol on PC keyboard or virtual keycode if negative
	eSamKey	nKey1, nKey2;		// Up to 2 SAM keys needed to generate the above symbol
	BYTE	bScanCode, bShift;	// Raw scan code and shift state needed on PC for above symbol
}
COMBINATION_KEY;


BYTE abKeyStates[256];
inline bool IsPressed(BYTE bKey_)	{ return (abKeyStates[bKey_] & 0x80) != 0; }
inline void PressKey(BYTE bKey_)	{ abKeyStates[bKey_] |= 0x80; }
inline void ReleaseKey(BYTE bKey_)	{ abKeyStates[bKey_] &= ~0x80; }
inline void ToggleKey(BYTE bKey_)	{ abKeyStates[bKey_] ^= 0x80; }

/*
// Mapping of SAM keyboard matrix positions to PC keys (simple non-symbol keys only)
static BYTE abSimpleKeys [SK_MAX] =
{
	DIK_LSHIFT,		DIK_Z,			DIK_X,		DIK_C,		DIK_V,		DIK_NUMPAD1,	DIK_NUMPAD2,	DIK_NUMPAD3,
	DIK_A,			DIK_S,			DIK_D,		DIK_F,		DIK_G,		DIK_NUMPAD4,	DIK_NUMPAD5,	DIK_NUMPAD6,
	DIK_Q,			DIK_W,			DIK_E,		DIK_R,		DIK_T,		DIK_NUMPAD7,	DIK_NUMPAD8,	DIK_NUMPAD9,
	DIK_1,			DIK_2,			DIK_3,		DIK_4,		DIK_5,		DIK_ESCAPE,		DIK_TAB,		DIK_CAPITAL,
	DIK_0,			DIK_9,			DIK_8,		DIK_7,		DIK_6,		0,				0,				DIK_BACK,
	DIK_P,			DIK_O,			DIK_I,		DIK_U,		DIK_Y,		0,				0,				DIK_NUMPAD0,
	DIK_RETURN,		DIK_L,			DIK_K,		DIK_J,		DIK_H,		0,				0,				DIK_RMENU,
	DIK_SPACE,		DIK_LCONTROL,	DIK_M,		DIK_N,		DIK_B,		0,				0,				DIK_INSERT,
	DIK_RCONTROL,	DIK_UP,			DIK_DOWN,	DIK_LEFT,	DIK_RIGHT
};

// Symbols with SAM keyboard details
COMBINATION_KEY asSamKeys[] =
{
	{ -DIK_DELETE,	SK_SHIFT, SK_DELETE },
	{ -DIK_HOME,	SK_CONTROL, SK_LEFT },
	{ -DIK_END,		SK_CONTROL, SK_RIGHT },
	{ -DIK_PRIOR,	SK_F4, SK_NONE },
	{ -DIK_NEXT,	SK_F1, SK_NONE },

	{ '!',	SK_SHIFT, SK_1 },		{ '@',	SK_SHIFT, SK_2 },		{ '#',	SK_SHIFT, SK_3 },
	{ '$',	SK_SHIFT, SK_4 },		{ '%',	SK_SHIFT, SK_5 },		{ '&',	SK_SHIFT, SK_6 },
	{ '\'', SK_SHIFT, SK_7 },		{ '(',	SK_SHIFT, SK_8 },		{ ')',	SK_SHIFT, SK_9 },
	{ '~',	SK_SHIFT, SK_0 },		{ '-',	SK_MINUS, SK_NONE },	{ '/',	SK_SHIFT, SK_MINUS },
	{ '+',	SK_PLUS, SK_NONE },		{ '*',	SK_SHIFT, SK_PLUS },	{ '<',	SK_SYMBOL, SK_Q },
	{ '>',	SK_SYMBOL, SK_W },		{ '[',	SK_SYMBOL, SK_R },		{ ']',	SK_SYMBOL, SK_T },
	{ '=',	SK_EQUALS, SK_NONE },	{ '_',	SK_SHIFT, SK_EQUALS },	{ '"',	SK_QUOTES, SK_NONE },
	{ '`',	SK_SHIFT, SK_QUOTES },	{ '{',	SK_SYMBOL, SK_F },		{ '}',	SK_SYMBOL, SK_G },
	{ '^',	SK_SYMBOL, SK_H },		{ (BYTE)'', SK_SYMBOL, SK_L },	{ ';',	SK_SEMICOLON, SK_NONE },
	{ ':',	SK_COLON, SK_NONE },	{ '?',	SK_SYMBOL, SK_X },		{ '.',	SK_PERIOD, SK_NONE },
	{ ',',	SK_COMMA, SK_NONE },	{ '\\',	SK_SHIFT, SK_INV },		{ '|',	SK_SYMBOL, SK_L },
	{ '\0' }
};

// Symbols with Spectrum keyboard details
COMBINATION_KEY asSpectrumKeys[] =
{
	{ '!',	SK_SYMBOL, SK_1 },		{ '@',	SK_SYMBOL, SK_2 },		{ '#',	SK_SYMBOL, SK_3 },
	{ '$',	SK_SYMBOL, SK_4 },		{ '%',	SK_SYMBOL, SK_5 },		{ '&',	SK_SYMBOL, SK_6 },
	{ '\'', SK_SYMBOL, SK_7 },		{ '(',	SK_SYMBOL, SK_8 },		{ ')',	SK_SYMBOL, SK_9 },
	{ '_',	SK_SYMBOL, SK_0 },		{ '<',	SK_SYMBOL, SK_R },		{ '>',	SK_SYMBOL, SK_T },
	{ ';',	SK_SYMBOL, SK_O },		{ '"',	SK_SYMBOL, SK_P },		{ '-',	SK_SYMBOL, SK_J },
	{ '^',	SK_SYMBOL, SK_H },		{ '+',	SK_SYMBOL, SK_K },		{ '=',	SK_SYMBOL, SK_L },
	{ ':',	SK_SYMBOL, SK_Z },		{ '',	SK_SYMBOL, SK_X },		{ '?',	SK_SYMBOL, SK_C },
	{ '/',	SK_SYMBOL, SK_V },		{ '*',	SK_SYMBOL, SK_B },		{ ',',	SK_SYMBOL, SK_N },
	{ '.',	SK_SYMBOL, SK_M },		{ '\b',	SK_SHIFT, SK_0 },
	{ '\0' }
};
*/

////////////////////////////////////////////////////////////////////////////////

static bool InitKeyboard ();
static bool InitMouse ();
static bool InitJoystick ();


bool Init ()
{
	Exit();

	InitJoystick();
	InitKeyboard();

	Mouse::Init();
	if (GetOption(mouse))
		InitMouse();

	return true;
}


void Exit ()
{
}


bool InitKeyboard ()
{
//	PrepareKeyTable(asSamKeys);
//	PrepareKeyTable(asSpectrumKeys);

	return true;
}

bool InitMouse ()
{
	return true;
}

bool InitJoystick ()
{
	// Remove any existing joysticks from the menu
//	while (DeleteMenu(hmenuJoysticks2, 1, MF_BYPOSITION) && DeleteMenu(hmenuJoysticks2, 1, MF_BYPOSITION));

	SetOption(joystick1, 0);
	SetOption(joystick2, 0);

	return true;
}

void UpdateAcquire (bool fActive_)
{
}

void ReadKeyboard ()
{
}

void ReadMouse ()
{
}


void ReadJoystick (int nDevice_, int nBaseKey_)
{
	// Simulate the PC key presses for the joystick movement and button(s?)
//	if (dijs.lX < -GetOption(joysticktol)) PressKey(nBaseKey_ + 0);	// Left
//	if (dijs.lX >  GetOption(joysticktol)) PressKey(nBaseKey_ + 1);	// Right
//	if (dijs.lY >  GetOption(joysticktol)) PressKey(nBaseKey_ + 2);	// Down
//	if (dijs.lY < -GetOption(joysticktol)) PressKey(nBaseKey_ + 3);	// Up
//	if (dijs.rgbButtons[0] & 0x80) PressKey(nBaseKey_ + 4);			// Fire
}


// Build the SAM keyboard matrix from the current PC state
void SetSamKeyState ()
{
	// No SAM keys are pressed initially
	ReleaseAllSamKeys();
}


void Update ()
{
	if (g_fActive)
	{
		// Read the devices that are present
//		if (pdiKeyboard) ReadKeyboard();
//		if (pdiMouse && GetOption(mouse)) ReadMouse();
//		if (apdiJoysticks[GetOption(joystick1)]) ReadJoystick(GetOption(joystick1), DIK_6);
//		if (apdiJoysticks[GetOption(joystick2)]) ReadJoystick(GetOption(joystick2), DIK_1);

		// Update the SAM keyboard matrix from the current key state (including joystick movement)
		SetSamKeyState();
	}
}

}	// namespace Input
