//
//	Win32 specific sound implementation using DirectSound to play the data produced by SAASOUND.DLL
//

#include "stdafx.h"

#include "SAM.h"
#include "CPU.h"
#include "Sound.h"
#include "Options.h"
#include "Util.h"
#include "SAASound.h"
#include "Profile.h"

// Hmmm... I don't like externs, so it might be worth using inline functions to retrieve these instead
extern int g_nLine, g_nLineCycle;


namespace Sound
{
// Offset of next sound write in streaming secondary buffer
static DWORD dwWriteOffset;

// Various values calculated from the sound settings
static UINT uSampleBufferSize, uSamplesPerFrame, uSamplesThisFrame, uBytesPerSample;

// Buffer used to build up fast sound changes within a frame
static BYTE* pbFrameSample;
static UINT uFrameOffset;

// Class pointer for Dave Hooper's SAA1099 emulation DLL
//static LPCSAASOUND pSAASound;

////////////////////////////////////////////////////////////////////////////////

bool Init ()
{
	// There's nothing to do if sound is disabled
	if (!GetOption(sound))
	{
		TRACE("Sound disabled, nothing to initialise\n");
		Exit();
		return true;
	}

	// Part close down so we can set things up again
	Exit(true);


	// Real implementation goes here


	// Clean up anything that was initialised, and disable the sound option
	Message(msgWarning, "Sound initialisation failed, disabling...");
	Exit();
	SetOption(sound,false);

	return false;
}

void Exit (bool fReinit_/*=false*/)
{
}

void Silence ()
{
	if (!GetOption(sound))
		return;
}

static DWORD GetSpaceAvailable ()
{
	return 0;
}

static void AddData (BYTE* pbSampleData_, UINT uSampleData_)
{
	DWORD dwSpace;

	// Return if no data was supplied, or there's no space available
	if (!uSampleData_ || !(dwSpace = GetSpaceAvailable()))
		return;
}


void Stop ()
{
}

void Start ()
{
}


void WriteAddress (BYTE bReg_)
{
//	if (pSAASound)
	{
		// Update the sample data with current values, then make the address change
		Update();
//		pSAASound->WriteAddress(bReg_ & 0x1f);
	}
}

void WriteData (BYTE bVal_)
{
//	if (pSAASound)
	{
		// Update the sample data with current values, then make the register change
		Update();
//		pSAASound->WriteData(bVal_);
	}
}

void Click (int bClick_)
{
//	if (pSAASound)
	{
		// Update the sample data with current values, then make the click
		Update();
//		pSAASound->ClickClick(bClick_);
	}
}

void FrameUpdate ()
{
	return;

//	if (pSAASound)
	{
		ProfileStart(Sound);

		// Add any buffered up sample data from the frames changes
		if (uFrameOffset)
		{
			AddData(pbFrameSample, uFrameOffset);
			uFrameOffset = 0;
		}

		// Generate the extra samples necessary to top up the sound buffer
		UINT uSamples = GetSpaceAvailable() / uBytesPerSample;
		if (uSamples)
		{
//			pSAASound->GenerateMany(pbFrameSample, uSamples);
			AddData(pbFrameSample, uSamples*uBytesPerSample);
		}

		// Reset the sample counter and frame offset for the next frame
		uSamplesThisFrame = 0;
		uFrameOffset = 0;

		ProfileEnd();
	}
}

void Update ()
{
	return;

//	if (pSAASound)
	{
		ProfileStart(Sound);

		// Calculate the number of cycles since the start of the current frame, and the number of
		// samples required to cover that period
		int nCurrentCycles = (g_nLine * TSTATES_PER_LINE) + g_nLineCycle;
		int uSamplesSoFar = (DWORD(uSamplesPerFrame) * DWORD(nCurrentCycles) / TSTATES_PER_FRAME);

		// Calculate the number of additional samples needed on top of what we've already generated
		int uSamples = uSamplesSoFar - uSamplesThisFrame;

		if (uSamples)
		{
			// Generate and appent the the additional sample(s) to our temporary buffer
//			pSAASound->GenerateMany(pbFrameSample + uFrameOffset, uSamples);
			uFrameOffset += uSamples * uBytesPerSample;

			// Update the total number of samples generated for the current frame
			uSamplesThisFrame = uSamplesSoFar;
		}

		ProfileEnd();
	}
}

};	// namespace Sound
