//	DirectX keyboard and joystick scanning functions
//
//	Notes:
//		Still seems to be a problem scanning Right-Alt on W2K - seems ok on
//		my Win98 tho.  May use Win32 API for that as an exception if necessary.
//
//		On startup (or detected keyboard change) two tables are build for the
//		keyboard mappings, to cope with different keyboard layouts for both
//		symbols (e.g. US) and letters (e.g. French).
//
//		I may well change to do handle keyboard change events instead of polling
//		for changes.

#include "stdafx.h"
#include <dinput.h>

#include "Input.h"
#include "IO.h"
#include "Util.h"
#include "Options.h"
#include "Mouse.h"
#include "UI.h"

extern HWND g_hwnd;
extern bool g_fActive;
extern BYTE keyports[9];

namespace Input
{
#define MAX_JOYSTICKS				8
#define KEYBOARD_BUFFER_SIZE		10

typedef struct
{
	int		nChar;				// Symbol on PC keyboard or scan-code if negative
	eSamKey	nKey1, nKey2;		// Up to 2 SAM keys needed to generate the above symbol
	BYTE	bScanCode, bShift;	// Raw scan code and shift state needed on PC for above symbol
}
COMBINATION_KEY;

typedef struct
{
	int		nChar;
	BYTE	bScanCode;
}
SIMPLE_KEY;

LPDIRECTINPUT pdi;
LPDIRECTINPUTDEVICE	pdiMouse, pdiKeyboard;

int nNextJoystick;
LPDIRECTINPUTDEVICE2 apdiJoysticks[MAX_JOYSTICKS];

BYTE abKeyStates[256];
inline bool IsPressed(BYTE bKey_)	{ return (abKeyStates[bKey_] & 0x80) != 0; }
inline void PressKey(BYTE bKey_)	{ abKeyStates[bKey_] |= 0x80; }
inline void ReleaseKey(BYTE bKey_)	{ abKeyStates[bKey_] &= ~0x80; }
inline void ToggleKey(BYTE bKey_)	{ abKeyStates[bKey_] ^= 0x80; }


// Mapping of SAM keyboard matrix positions to single PC keys.  Before use, these are converted to the corresponding
// DirectInput scan-codes for the current input locale, and are updated if the keyboard layout is changed.
SIMPLE_KEY asSamKeys [SK_MAX] =
{
	{-DIK_LSHIFT},	{'Z'},			{'X'},		{'C'},		{'V'},		{-DIK_NUMPAD1},	{-DIK_NUMPAD2},	{-DIK_NUMPAD3},
	{'A'},			{'S'},			{'D'},		{'F'},		{'G'},		{-DIK_NUMPAD4},	{-DIK_NUMPAD5},	{-DIK_NUMPAD6},
	{'Q'},			{'W'},			{'E'},		{'R'},		{'T'},		{-DIK_NUMPAD7},	{-DIK_NUMPAD8},	{-DIK_NUMPAD9},
	{'1'},			{'2'},			{'3'},		{'4'},		{'5'},		{-DIK_ESCAPE},	{-DIK_TAB},		{-DIK_CAPITAL},
	{'0'},			{'9'},			{'8'},		{'7'},		{'6'},		{0},			{0},			{-DIK_BACK},
	{'P'},			{'O'},			{'I'},		{'U'},		{'Y'},		{0},			{0},			{-DIK_NUMPAD0},
	{-DIK_RETURN},	{'L'},			{'K'},		{'J'},		{'H'},		{0},			{0},			{-DIK_RMENU},
	{-DIK_SPACE},	{-DIK_LCONTROL},{'M'},		{'N'},		{'B'},		{0},			{0},			{-DIK_INSERT},
	{-DIK_RCONTROL},{-DIK_UP},		{-DIK_DOWN},{-DIK_LEFT},{-DIK_RIGHT}
};

// Symbols with SAM keyboard details
COMBINATION_KEY asSamSymbols[] =
{
	{ -DIK_DELETE,	SK_SHIFT, SK_DELETE },
	{ -DIK_HOME,	SK_CONTROL, SK_LEFT },
	{ -DIK_END,		SK_CONTROL, SK_RIGHT },
	{ -DIK_PRIOR,	SK_F4, SK_NONE },
	{ -DIK_NEXT,	SK_F1, SK_NONE },
	{ -DIK_NUMLOCK,	SK_SYMBOL, SK_EDIT },

	{ '!',	SK_SHIFT, SK_1 },		{ '@',	SK_SHIFT, SK_2 },		{ '#',	SK_SHIFT, SK_3 },
	{ '$',	SK_SHIFT, SK_4 },		{ '%',	SK_SHIFT, SK_5 },		{ '&',	SK_SHIFT, SK_6 },
	{ '\'', SK_SHIFT, SK_7 },		{ '(',	SK_SHIFT, SK_8 },		{ ')',	SK_SHIFT, SK_9 },
	{ '~',	SK_SHIFT, SK_0 },		{ '-',	SK_MINUS, SK_NONE },	{ '/',	SK_SHIFT, SK_MINUS },
	{ '+',	SK_PLUS, SK_NONE },		{ '*',	SK_SHIFT, SK_PLUS },	{ '<',	SK_SYMBOL, SK_Q },
	{ '>',	SK_SYMBOL, SK_W },		{ '[',	SK_SYMBOL, SK_R },		{ ']',	SK_SYMBOL, SK_T },
	{ '=',	SK_EQUALS, SK_NONE },	{ '_',	SK_SHIFT, SK_EQUALS },	{ '"',	SK_QUOTES, SK_NONE },
	{ '`',	SK_SHIFT, SK_QUOTES },	{ '{',	SK_SYMBOL, SK_F },		{ '}',	SK_SYMBOL, SK_G },
	{ '^',	SK_SYMBOL, SK_H },		{ (BYTE)'', SK_SYMBOL, SK_L },	{ ';',	SK_SEMICOLON, SK_NONE },
	{ ':',	SK_COLON, SK_NONE },	{ '?',	SK_SYMBOL, SK_X },		{ '.',	SK_PERIOD, SK_NONE },
	{ ',',	SK_COMMA, SK_NONE },	{ '\\',	SK_SHIFT, SK_INV },		{ '|',	SK_SYMBOL, SK_L },
	{ '\0' }
};

// Symbols with Spectrum keyboard details
COMBINATION_KEY asSpectrumSymbols[] =
{
	{ '!',	SK_SYMBOL, SK_1 },		{ '@',	SK_SYMBOL, SK_2 },		{ '#',	SK_SYMBOL, SK_3 },
	{ '$',	SK_SYMBOL, SK_4 },		{ '%',	SK_SYMBOL, SK_5 },		{ '&',	SK_SYMBOL, SK_6 },
	{ '\'', SK_SYMBOL, SK_7 },		{ '(',	SK_SYMBOL, SK_8 },		{ ')',	SK_SYMBOL, SK_9 },
	{ '_',	SK_SYMBOL, SK_0 },		{ '<',	SK_SYMBOL, SK_R },		{ '>',	SK_SYMBOL, SK_T },
	{ ';',	SK_SYMBOL, SK_O },		{ '"',	SK_SYMBOL, SK_P },		{ '-',	SK_SYMBOL, SK_J },
	{ '^',	SK_SYMBOL, SK_H },		{ '+',	SK_SYMBOL, SK_K },		{ '=',	SK_SYMBOL, SK_L },
	{ ':',	SK_SYMBOL, SK_Z },		{ '',	SK_SYMBOL, SK_X },		{ '?',	SK_SYMBOL, SK_C },
	{ '/',	SK_SYMBOL, SK_V },		{ '*',	SK_SYMBOL, SK_B },		{ ',',	SK_SYMBOL, SK_N },
	{ '.',	SK_SYMBOL, SK_M },		{ '\b',	SK_SHIFT,  SK_0 },
	{ '\0' }
};

////////////////////////////////////////////////////////////////////////////////

static bool InitKeyboard ();
static bool InitMouse ();
static bool InitJoystick ();
static void PrepareKeyTable (COMBINATION_KEY* asKeys_);
static void PrepareKeyTable (SIMPLE_KEY* asKeys_);


bool Init ()
{
	Exit();

	// If we can find DirectInput 5.0 we can have joystick support
	HRESULT hr;
	if (SUCCEEDED(hr = DirectInputCreate(__hinstance, 0x0500, &pdi, NULL)))
		InitJoystick();

	// Otherwise just try for 3.0
	else if (FAILED(hr = DirectInputCreate(__hinstance, 0x0300, &pdi, NULL)))
		return false;

	InitKeyboard();

	Mouse::Init();
	if (GetOption(mouse))
		InitMouse();

	return true;
}


void Exit ()
{
	if (pdiKeyboard) { pdiKeyboard->Unacquire(); pdiKeyboard->Release(); pdiKeyboard = NULL; }
	if (pdiMouse) { pdiMouse->Unacquire(); pdiMouse->Release(); pdiMouse = NULL; }

	for (int i = 0 ; i < MAX_JOYSTICKS ; i++)
	{
		if (apdiJoysticks[i])
		{
			apdiJoysticks[i]->Unacquire();
			apdiJoysticks[i]->Release();
			apdiJoysticks[i] = NULL;
		}
	}

	if (pdi) { pdi->Release(); pdi = NULL; }

	Mouse::Exit();
}


bool InitKeyboard ()
{
	PrepareKeyTable(asSamSymbols);
	PrepareKeyTable(asSpectrumSymbols);
	PrepareKeyTable(asSamKeys);

	HRESULT hr;
	if (FAILED(hr = pdi->CreateDevice(GUID_SysKeyboard, &pdiKeyboard, NULL)))
		TRACE("!!! Failed to create keyboard device (%#08lx)\n", hr);
	else if (FAILED(hr = pdiKeyboard->SetCooperativeLevel(g_hwnd, DISCL_NONEXCLUSIVE | DISCL_FOREGROUND)))
		TRACE("!!! Failed to set cooperative level of keyboard device (%#08lx)\n", hr);
	else if (FAILED(hr = pdiKeyboard->SetDataFormat(&c_dfDIKeyboard)))
		TRACE("!!! Failed to set data format of keyboard device (%#08lx)\n", hr);
	else if (FAILED(hr = pdiKeyboard->Acquire()))
		TRACE("!!! Failed to set acquire keyboard device (%#08lx)\n", hr);
/*
	else
	{
		DIPROPDWORD dipdw =
        {
            {
                sizeof DIPROPDWORD,		// diph.dwSize
                sizeof DIPROPHEADER,	// diph.dwHeaderSize
                0,						// diph.dwObj
                DIPH_DEVICE,			// diph.dwHow
            },
            KEYBOARD_BUFFER_SIZE,		// dwData
        };

		if (FAILED(hr = g_pKeyboard->SetProperty(DIPROP_BUFFERSIZE, &dipdw.diph)))
			TRACE("!!! Failed to set keyboard buffer size\n", hr);
*/
		else
			return true;
//	}

	return false;
}

bool InitMouse ()
{
	HRESULT hr;

	if (FAILED(hr = pdi->CreateDevice(GUID_SysMouse, &pdiMouse, NULL)))
		TRACE("!!! Failed to create mouse device (%#08lx)\n", hr);
	else if (FAILED(hr = pdiMouse->SetCooperativeLevel(g_hwnd, DISCL_EXCLUSIVE | DISCL_FOREGROUND)))
		TRACE("!!! Failed to set cooperative level of mouse device (%#08lx)\n", hr);
	else if (FAILED(hr = pdiMouse->SetDataFormat(&c_dfDIMouse)))
		TRACE("!!! Failed to set data format of mouse device (%#08lx)\n", hr);
	else if (FAILED(hr = pdiMouse->Acquire()))
		TRACE("!!! Failed to set acquire mouse device (%#08lx)\n", hr);
	else
		return true;

	return false;
}

BOOL CALLBACK EnumJoystickProc (LPCDIDEVICEINSTANCE pdiDevice_, LPVOID lpv_)
{
	HRESULT hr;

	LPDIRECTINPUTDEVICE	pdiJoystick;
	if (FAILED(hr = pdi->CreateDevice(pdiDevice_->guidInstance, &pdiJoystick, NULL)))
		TRACE("!!! Failed to create joystick device (%#08lx)\n", hr);
	else
	{
		IDirectInputDevice2* pDevice;
		hr = pdiJoystick->QueryInterface(IID_IDirectInputDevice2, reinterpret_cast<void **>(&pDevice));
		pdiJoystick->Release();

		if (FAILED(hr))
			TRACE("!!! Failed to query joystick for IID_IDirectInputDevice2 (%#08lx)\n", hr);
		else
		{
			if (FAILED(pDevice->SetDataFormat(&c_dfDIJoystick)))
				TRACE("!!! Failed to set data format of joystick device (%#08lx)\n", hr);
			else
			{
				DIPROPRANGE diPrg;

				diPrg.diph.dwSize		=	sizeof diPrg;
				diPrg.diph.dwHeaderSize	=	sizeof diPrg.diph;
				diPrg.diph.dwObj		=	0;
				diPrg.diph.dwHow		=	DIPH_DEVICE;
				diPrg.lMin				=	-100;
				diPrg.lMax				=	+100;

				if (FAILED(pDevice->SetProperty(DIPROP_RANGE, &diPrg.diph)))
					TRACE("!!! Failed to set joystick range property (%#08lx)\n", hr);
				else
				{
					DIDEVICEINSTANCE didi;
					didi.dwSize = sizeof didi;
					pDevice->GetDeviceInfo(&didi);

					AppendMenu(hmenuJoysticks1, MF_STRING, IDM_HARDWARE_JOYSTICK1_NONE+nNextJoystick, didi.tszProductName);
					AppendMenu(hmenuJoysticks2, MF_STRING, IDM_HARDWARE_JOYSTICK2_NONE+nNextJoystick, didi.tszProductName);

					apdiJoysticks[nNextJoystick++] = pDevice;
					return (nNextJoystick == MAX_JOYSTICKS) ? DIENUM_STOP : DIENUM_CONTINUE;
				}
			}

			// Wait to set something up so release the device
			pDevice->Release();
		}
	}

	// Continue looking for other devices, even tho we failed with the current one
	return DIENUM_CONTINUE;
}

bool InitJoystick ()
{
	HRESULT hr;

	// Remove any existing joysticks from the menu
	while (DeleteMenu(hmenuJoysticks2, 1, MF_BYPOSITION) && DeleteMenu(hmenuJoysticks2, 1, MF_BYPOSITION));

	// Clear the joystick array
	ZeroMemory(&apdiJoysticks, sizeof apdiJoysticks);
	nNextJoystick = 1;

	// Enumerate the joystick devices
	if (FAILED(hr = pdi->EnumDevices(DIDEVTYPE_JOYSTICK, EnumJoystickProc, NULL, DIEDFL_ATTACHEDONLY)))
		TRACE("!!! Failed to enumerate joystick devices (%#08lx)\n", hr);

	if (GetOption(joystick1) > MAX_JOYSTICKS || !apdiJoysticks[GetOption(joystick1)])
		SetOption(joystick1, 0);

	if (GetOption(joystick2) > MAX_JOYSTICKS || !apdiJoysticks[GetOption(joystick2)])
		SetOption(joystick2, 0);

	return true;
}

void UpdateAcquire (bool fActive_)
{
	if (fActive_)
	{
		if (pdiKeyboard) pdiKeyboard->Acquire();
		if (pdiMouse) pdiMouse->Acquire();

		if (GetOption(joystick1) && apdiJoysticks[GetOption(joystick1)])
			apdiJoysticks[GetOption(joystick1)]->Acquire();

		if (GetOption(joystick2) && apdiJoysticks[GetOption(joystick2)])
			apdiJoysticks[GetOption(joystick2)]->Acquire();
	}
	else
	{
		if (pdiKeyboard) pdiKeyboard->Unacquire();
		if (pdiMouse) pdiMouse->Unacquire();

		if (GetOption(joystick1) && apdiJoysticks[GetOption(joystick1)])
			apdiJoysticks[GetOption(joystick1)]->Unacquire();

		if (GetOption(joystick2) && apdiJoysticks[GetOption(joystick2)])
			apdiJoysticks[GetOption(joystick2)]->Unacquire();
	}
}

void ReadKeyboard ()
{
	// Read the full PC keyboard, re-acquiring it if necessary
	HRESULT hr;
	if (FAILED(hr = pdiKeyboard->GetDeviceState(sizeof abKeyStates, abKeyStates)) &&
		SUCCEEDED(hr = pdiKeyboard->Acquire()) &&
		FAILED(hr = pdiKeyboard->GetDeviceState(sizeof abKeyStates, abKeyStates)))
	{
		TRACE("!!! Failed to read keyboard state! (%#08lx)\n", hr);
		ZeroMemory(abKeyStates, sizeof abKeyStates);
	}
	else
	{
		// If the option is set, Left-ALT does the same as Right-Control, to generate SAM Cntrl
		if (GetOption(altforcntrl) && IsPressed(DIK_LMENU))
			PressKey(DIK_RCONTROL);

		// Right-ALT is sometimes seen as Ctrl-ALT, so release Left-Ctrl (used for Symbol) if so
		if (IsPressed(DIK_RMENU))
			ReleaseKey(DIK_LCONTROL);
	}
}

void ReadMouse ()
{
	HRESULT hr;

	DIMOUSESTATE dims;
	ZeroMemory(&dims, sizeof dims);

	if (FAILED(hr = pdiMouse->GetDeviceState(sizeof dims, &dims)))
	{
		if (SUCCEEDED(hr = pdiMouse->Acquire()) && FAILED(hr = pdiMouse->GetDeviceState(sizeof dims, &dims)))
			TRACE("!!! Failed to read mouse state! (%#08lx)\n", hr);
	}

	// If something failed clear the key state table so nothing is pressed
	if (dims.lX || dims.lY)
		Mouse::Move(dims.lX, -dims.lY);

	for (int i = 0 ; i < sizeof dims.rgbButtons ; i++)
		Mouse::SetButton(i+1, (dims.rgbButtons[i] & 0x80) != 0);
}


void ReadJoystick (int nDevice_, int nBaseKey_)
{
	HRESULT hr;

	DIJOYSTATE dijs;
	ZeroMemory(&dijs, sizeof dijs);

	if (apdiJoysticks[nDevice_] && (FAILED(hr = apdiJoysticks[nDevice_]->Poll()) ||
		FAILED(hr = apdiJoysticks[nDevice_]->GetDeviceState(sizeof dijs, &dijs))))
	{
		if (SUCCEEDED(hr = apdiJoysticks[nDevice_]->Acquire()) && (FAILED(hr = apdiJoysticks[nDevice_]->Poll()) ||
			FAILED(hr = apdiJoysticks[nDevice_]->GetDeviceState(sizeof dijs, &dijs))))
			TRACE("!!! Failed to read joystick state! (%#08lx)\n", hr);
	}

	// Simulate the PC key presses for the joystick movement and button(s?)
	if (dijs.lX < -GetOption(joysticktol)) PressKey(nBaseKey_ + 0);	// Left
	if (dijs.lX >  GetOption(joysticktol)) PressKey(nBaseKey_ + 1);	// Right
	if (dijs.lY >  GetOption(joysticktol)) PressKey(nBaseKey_ + 2);	// Down
	if (dijs.lY < -GetOption(joysticktol)) PressKey(nBaseKey_ + 3);	// Up
	if (dijs.rgbButtons[0] & 0x80) PressKey(nBaseKey_ + 4);			// Fire
}


// Prepare the more complicated combination keys that are fairly keyboard specific
void PrepareKeyTable (SIMPLE_KEY* asKeys_)
{
	HKL hkl = GetKeyboardLayout(0);

	for (int i = 0 ; i < SK_MAX ; i++)
	{
		if (asKeys_[i].nChar)
			asKeys_[i].bScanCode = asKeys_[i].nChar > 0 ? MapVirtualKeyEx(asKeys_[i].nChar, 0, hkl) : -asKeys_[i].nChar;
	}
}

// Prepare the more complicated combination keys that are fairly keyboard specific
void PrepareKeyTable (COMBINATION_KEY* asKeys_)
{
	// Fetch the keyboard layout for the current thread
	HKL hkl = GetKeyboardLayout(0);

	for (int i = 0 ; asKeys_[i].nChar ; i++)
	{
		// Negative values are raw DirectInput key codes
		if (asKeys_[i].nChar < 0)
		{
			asKeys_[i].bScanCode = -asKeys_[i].nChar;
			asKeys_[i].bShift = 0;
		}

		// Convert the symbol to a virtual key sequence, and then to a keyboard scan-code
		else if (asKeys_[i].nChar)
		{
			WORD wVirtual = VkKeyScanEx(asKeys_[i].nChar, hkl);
			asKeys_[i].bScanCode = MapVirtualKeyEx(LOBYTE(wVirtual), 0, hkl);
			asKeys_[i].bShift = (wVirtual >> 8) & 1;
		}
	}
}

// Process more complicated key combinations
void ProcessKeyTable (COMBINATION_KEY* asKeys_)
{
	BYTE bShifted = IsPressed(DIK_LSHIFT);

	for (int i = 0 ; asKeys_[i].nChar ; i++)
	{
		int nSymbol = asKeys_[i].nChar;
		if (IsPressed(asKeys_[i].bScanCode) && !(asKeys_[i].bShift ^ (int)bShifted))
		{
			// Release the PC keys used for the key combination
			ReleaseKey(asKeys_[i].bScanCode);
			if (asKeys_[i].bShift) ToggleKey(DIK_LSHIFT);

			// Press 1 or 2 SAM keys to generate the symbol
			PressSamKey(asKeys_[i].nKey1);
			PressSamKey(asKeys_[i].nKey2);
		}
	}
}


// Build the SAM keyboard matrix from the current PC state
void SetSamKeyState ()
{
	// Return to ignore common Windows ALT- combinations so the SAM doesn't see them
	if (IsPressed(DIK_LMENU) && (IsPressed(DIK_TAB) || IsPressed(DIK_ESCAPE) || IsPressed(DIK_SPACE)))
		return;

	// No SAM keys are pressed initially
	ReleaseAllSamKeys();

	// Left and right shift keys are equivalent, and also complementary!
	bool fShiftToggle = IsPressed(DIK_LSHIFT) && IsPressed(DIK_RSHIFT);
	abKeyStates[DIK_LSHIFT]	|= abKeyStates[DIK_RSHIFT];
	abKeyStates[DIK_RCONTROL] |= abKeyStates[DIK_LMENU];

	// Process the key combinations required for the mode we're in
	switch (GetOption(keyboard))
	{
		case 0:											break;	// Raw keyboard
		case 1:		ProcessKeyTable(asSamSymbols);		break;	// SAM symbol keys
		case 2:		ProcessKeyTable(asSpectrumSymbols);	break;	// Spectrum symbol keys
	}

	// Toggle shift if both shift keys are down to allow shifted versions of keys that are
	// shifted on the PC but unshifted on the SAM
	if (fShiftToggle)
		ToggleKey(DIK_LSHIFT);

	// Build the rest of the SAM matrix from the simple non-symbol PC keys
	for (int i = 0 ; i < SK_MAX ; i++)
	{
		if (IsPressed(asSamKeys[i].bScanCode))
		{
			TRACE("Key pressed: %#04x\n", asSamKeys[i].bScanCode);
			PressSamKey(i);
		}
	}
}


void Update ()
{
	if (g_fActive)
	{
		// Read the devices that are present
		if (pdiKeyboard) ReadKeyboard();
		if (pdiMouse && GetOption(mouse)) ReadMouse();
		if (apdiJoysticks[GetOption(joystick1)]) ReadJoystick(GetOption(joystick1), DIK_6);
		if (apdiJoysticks[GetOption(joystick2)]) ReadJoystick(GetOption(joystick2), DIK_1);

		// Update the SAM keyboard matrix from the current key state (including joystick movement)
		SetSamKeyState();
	}
}

}	// namespace Input
