#ifndef SOUND_H
#define SOUND_H

#include "Options.h"
#include <dsound.h>

namespace Sound
{
	bool Init ();
	void Exit (bool fReinit_=false);

	void Out (WORD wPort_, BYTE bVal_);
	void Click (int bClick_);
	void FrameUpdate ();

	void Stop ();
	void Play ();
	void Silence ();
};


// Abstract base class to be used for all sound driver implementations
class CSoundDriver
{
	public:
		virtual ~CSoundDriver () { }

	public:
		virtual bool Init () = 0;

		virtual void Out (WORD wPort_, BYTE bVal_) = 0;
		virtual void Click (BYTE bVal) = 0;
		virtual void Update (bool fFrameEnd_=false) = 0;

		virtual bool Stop () = 0;
		virtual bool Play () = 0;
		virtual void Silence () = 0;
};


class CStreamingSound
{
	public:
		CStreamingSound (int nFreq_=0, int nBits_=0, int nChannels_=0);
		virtual ~CStreamingSound ();

	public:
		virtual bool Init ();
		virtual void Generate (BYTE* pb_, UINT uSamples_) = 0;
		virtual void GenerateExtra (BYTE* pb_, UINT uSamples_);

	public:
		virtual bool Play () { return m_pdsb && SUCCEEDED(m_pdsb->Play(0, 0, DSBPLAY_LOOPING)); }
		virtual bool Stop()  { Silence(); return m_pdsb && SUCCEEDED(m_pdsb->Stop()); }
		virtual void Silence ();

		virtual DWORD GetSpaceAvailable ();
		virtual void Update (bool fFrameEnd_=false);
		virtual void AddData (BYTE* pbSampleData_, UINT uSamples_);

	protected:
		IDirectSoundBuffer* m_pdsb;
		int m_nFreq, m_nBits, m_nChannels;
		UINT m_uSampleSize, m_uSamplesPerFrame, m_uSamplesThisFrame;

		DWORD m_dwWriteOffset;
		BYTE* m_pbFrameSample;
		UINT m_uSampleBufferSize, m_uFrameOffset;

		static int s_nUsage;
		static IDirectSound* s_pds;
		static IDirectSoundBuffer* s_pdsbPrimary;

	protected:
		bool InitDirectSound ();
		void ExitDirectSound ();
};


class CMonoDAC : public CStreamingSound
{
	public:
		CMonoDAC () : CStreamingSound(0, 8, 1), m_bSample(0x80) { }

	public:
		void Update (bool fFrameEnd_, BYTE bSample_)
		{
			CStreamingSound::Update(fFrameEnd_);
			m_bSample = bSample_;
		}

		void Generate (BYTE* pb_, UINT uSamples_)
		{
			memset(pb_, m_bSample, uSamples_);
		}

	public:
		BYTE m_bSample;
};


class CStereoDAC : public CStreamingSound
{
	public:
		CStereoDAC () : CStreamingSound(0, 8, 2), m_bLeft(0x80), m_bRight(0x80) { }

	public:
		void Generate (BYTE* pb_, UINT uSamples_)
		{
			WORD* pw = reinterpret_cast<WORD*>(pb_);
			WORD wSample = (static_cast<WORD>(m_bRight) << 8) | m_bLeft;

			while (uSamples_--)
				*pw++ = wSample;
		}

	public:
		BYTE m_bLeft, m_bRight;
};

#endif	// SOUND_H
