//	Notes:
//		Functional but primitive UI with a menu and more recently a couple of dialogue boxes
//
//	ToDo:
//		Finish dialogues, add New Disk dialogue, tidy up code and remove extra crap put
//		in for testing etc.
//
//		Better current directory management for disk open dialogue boxes, as it's a bit
//		annoying to be put back to My Documents!
//
//		Check screensaver activation doesn't break us (didn't used to but might after recent changes)

#include "stdafx.h"

#include "CPU.h"
#include "Options.h"
#include "Util.h"
#include "Mouse.h"
#include "FDC.h"
#include "Memory.h"
#include "Sound.h"
#include "Input.h"
#include "Video.h"
#include "Frame.h"
#include "CDrive.h"
#include "Version.h"

extern Z80Regs regs;
extern CDrive g_Drive1, g_Drive2;

extern int __argc;
extern char** __argv;
extern int main (int argc, char *argv[]);


#include "resource.h"	// For menu and dialogue box symbols

BOOL CALLBACK ImportExportDlgProc (HWND hdlg_, UINT uMsg_, WPARAM wParam_, LPARAM lParam_);
void CentreWindow (HWND hwnd_, HWND hwndParent_=NULL);


bool g_fActive = true;

HINSTANCE __hinstance;
HWND g_hwnd;
HACCEL g_hAccel;
HMENU hmenuJoysticks1, hmenuJoysticks2;
extern HINSTANCE __hinstance;

RECT g_rWindow;



static char szDiskFilters [] =
	"All images (*.DSK;*.SAD;*.SDF)\0*.DSK;*.SAD;*.SDF\0"
#ifdef USE_ZLIB
	"All archives (*.GZ;*.ZIP)\0*.GZ;*.ZIP\0"
#endif
	"All Files (*.*)\0*.*\0";

bool InitWindow ();



int WINAPI WinMain(HINSTANCE hinst_, HINSTANCE hinstPrev_, LPSTR pszCmdLine_, int nCmdShow_)
{
	__hinstance = hinst_;

	return main(__argc, __argv);
}



namespace UI
{

bool Init ()
{
	SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_ABOVE_NORMAL);
	return InitWindow();
}

void Exit ()
{
	if (g_hwnd)
	{
		// When we reach here during a normal shutdown the window will already have gone, so check first
		if (IsWindow(g_hwnd))
			DestroyWindow(g_hwnd);

		g_hwnd = NULL;
	}
}


// Check and process any incoming messages
bool CheckEvents ()
{
	while (1)
	{
		// Loop to process any pending Windows messages
		MSG msg;
		while (PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE))
		{
			// Return on WM_QUIT
			if (!GetMessage(&msg, NULL, 0, 0 ))
				return false;

			// Only do keyboard translation (so the menu shortcuts work) when Alt is not being used for Cntrl
			if (!TranslateAccelerator(g_hwnd, g_hAccel, &msg) && !GetOption(altforcntrl))
				TranslateMessage(&msg);

			// Dispatch the message
			DispatchMessage(&msg);

			// Exit early to process key changes (not including repeats)
			if (msg.message == WM_KEYUP || (msg.message == WM_KEYDOWN && !(msg.lParam & 0x40000000)))
				break;
		}

		// Break out to do more processing if we're active
		if (g_fActive)
			break;

		// Be system friendly and block until there's something to do
		WaitMessage();
	}

	return true;
}

void ShowMessage (eMsgType eType_, const char* pcszMessage_)
{
	switch (eType_)
	{
		case msgInfo:		break;
		case msgWarning:	MessageBox(NULL, pcszMessage_, "Warning", MB_OK | MB_ICONEXCLAMATION);	break;
		case msgError:		MessageBox(NULL, pcszMessage_, "Error", MB_OK | MB_ICONSTOP);			break;

		case msgFatal:
			MessageBox(NULL, pcszMessage_, "Fatal Error", MB_OK | MB_ICONSTOP);

			// Something went seriously wrong, so give the debugger a look in and abort!
			DebugBreak();
			abort();
			break;
	}
}

};


bool GetLoadFile (HWND hwndParent_, LPCSTR pcszFilters_, LPCSTR pcszDefExt_, LPSTR pcszFile_, DWORD cbSize_, bool* pfReadOnly_)
{
	char szFile [256] = "";

	// Fill in the details for a fax file to select
	OPENFILENAME ofn;
	ZeroMemory(&ofn, sizeof ofn);
	ofn.lStructSize		= sizeof ofn;
	ofn.hwndOwner		= hwndParent_;
	ofn.lpstrFilter		= pcszFilters_;
	ofn.lpstrFile		= pcszFile_;
	ofn.nMaxFile		= cbSize_;
	ofn.lpstrDefExt		= pcszDefExt_;
	ofn.Flags			= OFN_EXPLORER | OFN_FILEMUSTEXIST | OFN_NOCHANGEDIR;
	ofn.lpstrTitle		= "Select Disk Image";

	// If supplied and set, check the read-only checkbox
	if (pfReadOnly_ && *pfReadOnly_)
		ofn.Flags |= OFN_READONLY;

	// Return if the user cancelled without picking a file
	BOOL fRet = GetOpenFileName(&ofn);
	if (!fRet)
	{
		// Invalid paths can choke the dialog, so clear out the file and try again
		if (CommDlgExtendedError() == FNERR_INVALIDFILENAME)
		{
			*pcszFile_ = '\0';
			fRet = GetOpenFileName(&ofn);
		}
	}

	if (!fRet)
	{
		TRACE("GetOpenFileName() failed (%#08lx)\n", CommDlgExtendedError());
		return FALSE;
	}

	// If required, read the requested read-only state of the input file
	if (pfReadOnly_)
		*pfReadOnly_ = (ofn.Flags & OFN_READONLY) != 0;

	// Copy the filename into the supplied string, and return success
	TRACE("GetLoadFile: %s selected\n", pcszFile_);
	return TRUE;
}

bool GetSaveFile (HWND hwndParent_, LPCSTR pcszFilters_, LPCSTR pcszDefExt_, LPSTR pcszFile_, DWORD cbSize_)
{
	return FALSE;
}


void InsertDisk (CDrive& rDrive_)
{
	char szFile[_MAX_PATH] = "";
	bool fReadOnly = false;

	// Eject any current disk, and use the path and read-only status from it for the open dialogue box
	CDisk* pDisk = rDrive_.Eject(false);
	if (pDisk)
	{
		strcpy(szFile, pDisk->GetName());
		fReadOnly = pDisk->IsReadOnly();
		rDrive_.Insert(pDisk);
	}

	// Prompt for the a new disk to insert
	if (GetLoadFile(g_hwnd, szDiskFilters, NULL, szFile, sizeof szFile, &fReadOnly))
	{
		// Eject any previous disk, saving if necessary
		rDrive_.Eject();

		// Open the new disk (using the requested read-only mode), and insert it into the drive if successful
		if (pDisk = CDisk::Open(szFile, fReadOnly))
			rDrive_.Insert(pDisk);
		else
			Message(msgWarning, "Invalid disk image: %s", szFile);
	}
}


void ResizeWindow ()
{
	int nWidth = (VISIBLE_BLOCKS + SCREEN_BLOCKS + VISIBLE_BLOCKS) << 3;
	int nHeight = VISIBLE_LINES + SCREEN_LINES + VISIBLE_LINES;

	if (GetOption(ratio5_4))
		nWidth = MulDiv(nWidth, 5, 4);

	if (GetOption(fullscreen))
	{
		DWORD dwStyle = GetWindowStyle(g_hwnd);
		dwStyle = WS_POPUP|WS_VISIBLE;
		SetWindowLong(g_hwnd, GWL_STYLE, dwStyle);

		RECT rect = { 0, 0, GetSystemMetrics(SM_CXSCREEN), GetSystemMetrics(SM_CYSCREEN) };
		SetWindowPos(g_hwnd, HWND_TOPMOST, rect.left, rect.top, rect.right-rect.left, rect.bottom-rect.top, 0);
	}
	else
	{
		// Leave a maximised window as it is
		WINDOWPLACEMENT wp;
		wp.length = sizeof wp;
		if (!GetWindowPlacement(g_hwnd, &wp) || (wp.showCmd != SW_SHOWMAXIMIZED))
		{
			DWORD dwStyle = (GetWindowStyle(g_hwnd) & WS_VISIBLE) | WS_OVERLAPPEDWINDOW;
			SetWindowLong(g_hwnd, GWL_STYLE, dwStyle);

			// Adjust the window rectangle to give the client area size required for the main window
			RECT rect = { 0, 0, (nWidth * GetOption(scale)), (nHeight * GetOption(scale)) };
			AdjustWindowRectEx(&rect, GetWindowStyle(g_hwnd), GetMenu(g_hwnd) != NULL, GetWindowExStyle(g_hwnd));
			SetWindowPos(g_hwnd, HWND_NOTOPMOST, 0, 0, rect.right-rect.left, rect.bottom-rect.top, SWP_NOMOVE);

			// Get the actual client rectangle, now that the menu is in place
			RECT rClient;
			GetClientRect(g_hwnd, &rClient);

			// If the menu has wrapped the client area won't be big enough, so enlarge if necessary
			if ((rClient.bottom - rClient.top) < (nHeight * GetOption(scale)))
			{
				rect.bottom += GetSystemMetrics(SM_CYMENUSIZE);
				SetWindowPos(g_hwnd, HWND_NOTOPMOST, 0, 0, rect.right-rect.left, rect.bottom-rect.top, SWP_NOMOVE);
			}
		}
	}

	// Ensure the window is repainted
	InvalidateRect(g_hwnd, NULL, TRUE);
}


#define CheckOption(id,check)	CheckMenuItem(hmenu, (id), (check) ? MF_CHECKED : MF_UNCHECKED)
#define EnableItem(id,enable)	EnableMenuItem(hmenu, (id), (enable) ? MF_ENABLED : MF_GRAYED)

void UpdateMenuFromOptions ()
{
	int	i;
	HMENU hmenu = GetMenu(g_hwnd), hmenuFile = GetSubMenu(hmenu, 0);


	// Grey the sub-menu for disabled drives, and update the status/text of the other Drive 1 options
	EnableMenuItem(hmenuFile, 1, (GetOption(drive1) == 1) ? MF_ENABLED|MF_BYPOSITION : MF_GRAYED|MF_BYPOSITION);
	EnableItem(IDM_FILE_FLOPPY1_SAVE_CHANGES, g_Drive1.IsModified());

	CDisk* pDisk;
	char szEject[128];
	g_Drive1.Insert(pDisk = g_Drive1.Eject(false));
	lstrcat(lstrcat(lstrcpy(szEject, "&Eject "), pDisk ? strrchr(pDisk->GetName(), '\\')+1 : ""), "\tShift-F1");
	ModifyMenu(hmenu, IDM_FILE_FLOPPY1_EJECT, MF_STRING|(pDisk ? MF_ENABLED : MF_GRAYED), IDM_FILE_FLOPPY1_EJECT, szEject);
	CheckOption(IDM_FILE_FLOPPY1_DEVICE, pDisk && CFloppyStream::IsRecognised(pDisk->GetName()));
//	EnableItem(IDM_FILE_FLOPPY1_DEVICE, true);

	// Grey the sub-menu for disabled drives, and update the status/text of the other Drive 2 options
	EnableMenuItem(hmenuFile, 2, (GetOption(drive2) == 1) ? MF_ENABLED|MF_BYPOSITION : MF_GRAYED|MF_BYPOSITION);
	EnableItem(IDM_FILE_FLOPPY2_SAVE_CHANGES, g_Drive2.IsModified());

	g_Drive2.Insert(pDisk = g_Drive2.Eject(false));
	lstrcat(lstrcat(lstrcpy(szEject, "&Eject "), pDisk ? strrchr(pDisk->GetName(), '\\')+1 : ""), "\tShift-F2");
	ModifyMenu(hmenu, IDM_FILE_FLOPPY2_EJECT, MF_STRING|(pDisk ? MF_ENABLED : MF_GRAYED), IDM_FILE_FLOPPY2_EJECT, szEject);
	CheckOption(IDM_FILE_FLOPPY2_DEVICE, pDisk && CFloppyStream::IsRecognised(pDisk->GetName()));


	CheckOption(IDM_CPU_RESET, 0);
	CheckOption(IDM_CPU_FAST_RESET, GetOption(fastreset));
	CheckOption(IDM_CPU_NMI, 0);
	CheckOption(IDM_CPU_PAUSE, InterruptType == Z80_pause);
	CheckOption(IDM_CPU_PROFILE, GetOption(profile));


	CheckOption(IDM_SOUND_ENABLED, GetOption(sound));

	CheckOption(IDM_SOUND_RATE_11025, GetOption(rate)==11025);
	CheckOption(IDM_SOUND_RATE_22050, GetOption(rate)==22050);
	CheckOption(IDM_SOUND_RATE_44100, GetOption(rate)==44100);

	CheckOption(IDM_SOUND_BITS_8, GetOption(bits)==8);
	CheckOption(IDM_SOUND_BITS_16, GetOption(bits)==16);

	CheckOption(IDM_SOUND_STEREO, GetOption(stereo));

	CheckOption(IDM_SOUND_FILTER, GetOption(filter));

	CheckOption(IDM_SOUND_LATENCY_1, GetOption(latency)==2);
	CheckOption(IDM_SOUND_LATENCY_2, GetOption(latency)==3);
	CheckOption(IDM_SOUND_LATENCY_3, GetOption(latency)==4);
	CheckOption(IDM_SOUND_LATENCY_4, GetOption(latency)==5);
	CheckOption(IDM_SOUND_LATENCY_5, GetOption(latency)==6);
	CheckOption(IDM_SOUND_LATENCY_10, GetOption(latency)==11);
	CheckOption(IDM_SOUND_LATENCY_15, GetOption(latency)==16);
	CheckOption(IDM_SOUND_LATENCY_20, GetOption(latency)==21);
	CheckOption(IDM_SOUND_LATENCY_25, GetOption(latency)==26);


	CheckOption(IDM_VIDEO_ENABLED, GetOption(video));

	CheckOption(IDM_VIDEO_SYNC_NONE, GetOption(sync) == 0);
	CheckOption(IDM_VIDEO_SYNC_TIMER, GetOption(sync) == 1);
	CheckOption(IDM_VIDEO_SYNC_VSYNC, GetOption(sync) == 2);

	CheckOption(IDM_VIDEO_FRAMESKIP_AUTO, GetOption(frameskip)==0);
	CheckOption(IDM_VIDEO_FRAMESKIP_NONE, GetOption(frameskip)==1);
	CheckOption(IDM_VIDEO_FRAMESKIP_1, GetOption(frameskip)==2);
	CheckOption(IDM_VIDEO_FRAMESKIP_2, GetOption(frameskip)==3);
	CheckOption(IDM_VIDEO_FRAMESKIP_3, GetOption(frameskip)==4);

	CheckOption(IDM_VIDEO_ACCURATE_MODE3, GetOption(sourcesize)==2);

	CheckOption(IDM_VIDEO_SCALING_1X1, GetOption(scale)==1);
	CheckOption(IDM_VIDEO_SCALING_2X2, GetOption(scale)==2);
	CheckOption(IDM_VIDEO_SCALING_3X3, GetOption(scale)==3);

	CheckOption(IDM_VIDEO_FULLSCREEN, GetOption(fullscreen));

	CheckOption(IDM_VIDEO_DEPTH_8, GetOption(depth)==8);
	CheckOption(IDM_VIDEO_DEPTH_16, GetOption(depth)==16);
	CheckOption(IDM_VIDEO_DEPTH_32, GetOption(depth)==32);

	CheckOption(IDM_VIDEO_METHOD_SOFTWARE, GetOption(videomethod)==0);
	CheckOption(IDM_VIDEO_METHOD_SYSTEM, GetOption(videomethod)==1);
	CheckOption(IDM_VIDEO_METHOD_VIDEO, GetOption(videomethod)==2);
	CheckOption(IDM_VIDEO_METHOD_OVERLAY_SYSTEM_TO_VIDEO, GetOption(videomethod)==3);
	CheckOption(IDM_VIDEO_METHOD_OVERLAY_SYSTEM, GetOption(videomethod)==4);
	CheckOption(IDM_VIDEO_METHOD_OVERLAY_VIDEO, GetOption(videomethod)==5);

	CheckOption(IDM_VIDEO_DRAW_SURFACE, GetOption(drawmethod)==0);
	CheckOption(IDM_VIDEO_DRAW_PALETTE, GetOption(drawmethod)==1);

	CheckOption(IDM_VIDEO_RATIO_5_4, GetOption(ratio5_4));

	CheckOption(IDM_VIDEO_FLOPPY_LEDS, GetOption(floppyleds));


	CheckOption(IDM_HARDWARE_MAIN_MEMORY_256K, GetOption(mainmem)==256);
	CheckOption(IDM_HARDWARE_MAIN_MEMORY_512K, GetOption(mainmem)==512);

	CheckOption(IDM_HARDWARE_EXTERNAL_MEMORY_NONE, GetOption(externalmem)==0);
	CheckOption(IDM_HARDWARE_EXTERNAL_MEMORY_1MB, GetOption(externalmem)==1);
	CheckOption(IDM_HARDWARE_EXTERNAL_MEMORY_2MB, GetOption(externalmem)==2);
	CheckOption(IDM_HARDWARE_EXTERNAL_MEMORY_3MB, GetOption(externalmem)==3);
	CheckOption(IDM_HARDWARE_EXTERNAL_MEMORY_4MB, GetOption(externalmem)==4);

	CheckOption(IDM_HARDWARE_DRIVE1_NONE, GetOption(drive1)==0);
	CheckOption(IDM_HARDWARE_DRIVE1_FLOPPY, GetOption(drive1)==1);

	CheckOption(IDM_HARDWARE_DRIVE2_NONE, GetOption(drive2)==0);
	CheckOption(IDM_HARDWARE_DRIVE2_FLOPPY, GetOption(drive2)==1);
	CheckOption(IDM_HARDWARE_DRIVE2_ATOM, GetOption(drive2)==2);
#ifdef _DEBUG
//	EnableItem(IDM_HARDWARE_DRIVE2_ATOM, TRUE);
#endif

	CheckOption(IDM_HARDWARE_KEYBOARD_RAW, GetOption(keyboard)==0);
	CheckOption(IDM_HARDWARE_KEYBOARD_SAM, GetOption(keyboard)==1);
	CheckOption(IDM_HARDWARE_KEYBOARD_SPECTRUM, GetOption(keyboard)==2);

	CheckOption(IDM_HARDWARE_ALT_FOR_CNTRL, GetOption(altforcntrl)!=0);

	CheckOption(IDM_HARDWARE_MOUSE_NONE, GetOption(mouse)==0);
	CheckOption(IDM_HARDWARE_MOUSE_ENABLED, GetOption(mouse)==1);
	CheckOption(IDM_HARDWARE_MOUSE_ENABLED_MODE3, GetOption(mouse)==2);

	for (i = 0 ; i < 16 ; i++) CheckOption(IDM_HARDWARE_JOYSTICK1_NONE+i, GetOption(joystick1)==i);
	for (i = 0 ; i < 16 ; i++) CheckOption(IDM_HARDWARE_JOYSTICK2_NONE+i, GetOption(joystick2)==i);

	CheckOption(IDM_HARDWARE_JOYSTICKTOL_0, GetOption(joysticktol)==0);
	CheckOption(IDM_HARDWARE_JOYSTICKTOL_10, GetOption(joysticktol)==10);
	CheckOption(IDM_HARDWARE_JOYSTICKTOL_20, GetOption(joysticktol)==20);
	CheckOption(IDM_HARDWARE_JOYSTICKTOL_30, GetOption(joysticktol)==30);
	CheckOption(IDM_HARDWARE_JOYSTICKTOL_40, GetOption(joysticktol)==40);
	CheckOption(IDM_HARDWARE_JOYSTICKTOL_50, GetOption(joysticktol)==50);

	CheckOption(IDM_HARDWARE_CLOCK_SAMBUS, GetOption(sambus_clock));
	CheckOption(IDM_HARDWARE_CLOCK_DALLAS, GetOption(dallas_clock));

	CheckOption(IDM_HARDWARE_SYSTEM_SAM, GetOption(spectrum)==0);
	CheckOption(IDM_HARDWARE_SYSTEM_SPECTRUM, GetOption(spectrum)==1);

	CheckOption(IDM_HARDWARE_PARALLEL1_NONE, GetOption(parallel1)==0);
	CheckOption(IDM_HARDWARE_PARALLEL1_PRINTER, GetOption(parallel1)==1);
	EnableItem(IDM_HARDWARE_PARALLEL1_PRINTER, FALSE);
	CheckOption(IDM_HARDWARE_PARALLEL1_MONODAC, GetOption(parallel1)==2);
	CheckOption(IDM_HARDWARE_PARALLEL1_STEREODAC, GetOption(parallel1)==3);

	CheckOption(IDM_HARDWARE_PARALLEL2_NONE, GetOption(parallel2)==0);
	CheckOption(IDM_HARDWARE_PARALLEL2_PRINTER, GetOption(parallel2)==1);
	EnableItem(IDM_HARDWARE_PARALLEL2_PRINTER, FALSE);
	CheckOption(IDM_HARDWARE_PARALLEL2_MONODAC, GetOption(parallel2)==2);
	CheckOption(IDM_HARDWARE_PARALLEL2_STEREODAC, GetOption(parallel2)==3);
}

void UpdateOptionsFromMenu (WORD wId_)
{
	// Sound menu commands
	switch (wId_ & 0xff00)
	{
		// Sound menu
		case 0x8000:
		{
			switch (wId_ & 0xff)
			{
				case 0x00:	SetOption(sound, !GetOption(sound));	break;

				case 0x10:	SetOption(rate, 11025);					break;
				case 0x11:	SetOption(rate, 22050);					break;
				case 0x12:	SetOption(rate, 44100);					break;

				case 0x20:	SetOption(bits, 8);						break;
				case 0x21:	SetOption(bits, 16);					break;

				case 0x30:	SetOption(stereo, !GetOption(stereo));	break;

				case 0x40:	SetOption(filter, !GetOption(filter));	break;

				case 0x50:	SetOption(latency, 2);					break;
				case 0x51:	SetOption(latency, 3);					break;
				case 0x52:	SetOption(latency, 4);					break;
				case 0x53:	SetOption(latency, 5);					break;
				case 0x54:	SetOption(latency, 6);					break;
				case 0x55:	SetOption(latency, 11);					break;
				case 0x56:	SetOption(latency, 16);					break;
				case 0x57:	SetOption(latency, 21);					break;
				case 0x58:	SetOption(latency, 26);					break;
			}

			// Reinitialise the sound system
			Sound::Init();
			break;
		}

		// Video menu
		case 0x8100:
		{
			switch (wId_ & 0xf0)
			{
				case 0x00:	SetOption(video, !GetOption(video));			break;
				case 0x10:	SetOption(sync, wId_ & 3);						break;
				case 0x20:	SetOption(frameskip, wId_ & 7);					break;
				case 0x30:	SetOption(sourcesize, 3-GetOption(sourcesize));	break;
				case 0x40:	SetOption(scale, (wId_ & 3) + 1);				break;
				case 0x50:	SetOption(ratio5_4, !GetOption(ratio5_4));		break;
				case 0x60:	SetOption(fullscreen, !GetOption(fullscreen));	break;
				case 0x70:	SetOption(depth, 8 << (wId_ & 3));				break;
				case 0x80:	SetOption(videomethod, (wId_ & 7));				break;
				case 0x90:	SetOption(floppyleds, !GetOption(floppyleds));	break;
				case 0xa0:	SetOption(drawmethod, (wId_ & 1));				break;
			}

			switch (wId_ & 0xf0)
			{
				case 0x00:
					Frame::Clear();
					Frame::Redraw();
					break;

				case 0x30:
					if (GetOption(fullscreen))
						Sound::Silence();

					Video::Init();
					break;

				case 0x40:
				case 0x50:
					ResizeWindow();
					Video::UpdateOverlay();
					break;

				case 0x60:
					// If we're going fullscreen, remember the windowed mode window position
					if (GetOption(fullscreen))
						GetWindowRect(g_hwnd, &g_rWindow);
					else
						SetWindowPos(g_hwnd, HWND_TOPMOST, g_rWindow.left, g_rWindow.top, 0, 0, SWP_NOSIZE);

					Sound::Silence();
					Video::Init();
					break;

				case 0x70:
					if (GetOption(fullscreen))
					{
						Sound::Silence();
						Video::Init();
					}
					break;

				case 0x80:
				case 0xa0:
					if (GetOption(fullscreen))
						Sound::Silence();
					Video::Init();
			}
			break;
		}

		// Hardware menu
		case 0x8200:
		case 0x8300:
		{
			switch (wId_ & 0xff0)
			{
				case 0x210:	SetOption(mainmem, (wId_ & 1) ? 512 : 256);			break;
				case 0x220:	SetOption(externalmem, wId_ & 7);					break;
				case 0x230:	SetOption(drive1, wId_ & 1);						break;
				case 0x240:	SetOption(drive2, wId_ & 3);						break;
				case 0x250:														break;
				case 0x260:	SetOption(keyboard, wId_ & 3);						break;
				case 0x270:	SetOption(altforcntrl, !GetOption(altforcntrl));	break;
				case 0x280:	SetOption(mouse, wId_ & 3);							break;
				case 0x290:	SetOption(joystick1, wId_ & 0xf);					break;
				case 0x2a0:	SetOption(joystick2, wId_ & 0xf);					break;
				case 0x2b0:	SetOption(joysticktol, (wId_ & 0xf) * 10);			break;
				case 0x2c0:	SetOption(sambus_clock, !GetOption(sambus_clock));	break;
				case 0x2d0:	SetOption(dallas_clock, !GetOption(dallas_clock));	break;
				case 0x2e0:	SetOption(spectrum, wId_ & 1); SetOption(keyboard, 1 + (wId_ & 1));	break;
				case 0x2f0:	SetOption(parallel1, wId_ & 0xf);					break;
				case 0x300:	SetOption(parallel2, wId_ & 0xf);					break;
			}

			switch (wId_ & 0xff0)
			{
				case 0x280:
					Input::Init();	break;

				case 0x2f0:
				case 0x300:
					Sound::Init();	break;
			}
			break;
		}

		// CPU menu
		case 0x8400:
		{
			switch (wId_ & 0xf0)
			{
				case 0x00:	InterruptType = Z80_reset;						break;
				case 0x10:	SetOption(fastreset, !GetOption(fastreset));	break;
				case 0x20:	InterruptType = Z80_nmi;						break;

				case 0x30:
				{
					static BYTE bSaveIff1;

					if (InterruptType != Z80_pause)
					{
						bSaveIff1 = regs.IFF1;
						InterruptType = Z80_pause;
					}
					else
					{
						InterruptType = Z80_none;
						regs.IFF1 = bSaveIff1;
					}
				}
				break;

				case 0x40:	SetOption(profile, !GetOption(profile));		break;
			}
		}
	}
}

void CentreWindow (HWND hwnd_, HWND hwndParent_/*=NULL*/)
{
	// If a window isn't specified get the parent window
	if (!hwndParent_ || (IsIconic(hwndParent_) || !(hwndParent_ = GetParent(hwnd_))))
		hwndParent_ = GetDesktopWindow();

	// Get the rectangles of the window to be centred and the parent window
	RECT rWindow, rParent;
	GetWindowRect(hwnd_, &rWindow);
	GetWindowRect(hwndParent_, &rParent);

	// Work out the new position of the window
	int nX = rParent.left + ((rParent.right - rParent.left) - (rWindow.right - rWindow.left)) / 2;
	int nY = rParent.top + ((rParent.bottom - rParent.top) - (rWindow.bottom - rWindow.top)) / 2;

	// Move the window to its new position
	SetWindowPos(hwnd_, NULL, nX, nY, 0, 0, SWP_SHOWWINDOW|SWP_NOSIZE|SWP_NOZORDER);
}


BOOL CALLBACK AboutDlgProc (HWND hdlg_, UINT uMsg_, WPARAM wParam_, LPARAM lParam_)
{
	switch (uMsg_)
	{
		case WM_INITDIALOG:
			CentreWindow(hdlg_);
			return 1;

		case WM_COMMAND:
			EndDialog(hdlg_, 0);
			return 1;
	}

	return 0;
}


long CALLBACK WindowProc (HWND hwnd_, UINT uMsg_, WPARAM wParam_, LPARAM lParam_)
{
	static bool fClosing = false, fInMenu = false;

    switch(uMsg_)
    {
		case WM_CREATE:
			// Allow files to be dropped onto our main window
			DragAcceptFiles(hwnd_, TRUE);
			Video::UpdateOverlay();
			break;

		// Application close request
		case WM_CLOSE:
			// Flag we're closing and signal the message loop to end
			fClosing = true;
			PostQuitMessage(0);
			break;

		case WM_INPUTLANGCHANGE:
			Input::Init();
			return 1;

		// File has been dropped on our window
		case WM_DROPFILES:
		{
			char szFile[_MAX_PATH]="";

			// Insert the first (or only) dropped file into drive 1			
			if (DragQueryFile(reinterpret_cast<HDROP>(wParam_), 0, szFile, sizeof szFile))
				g_Drive1.Insert(szFile);

			return 0;
		}

		case WM_SYSCOMMAND:
			// Deactivate ourselves immediately if the screensaver kicks in
			if (wParam_ == SC_SCREENSAVE)
			{
				Sound::Silence();
				g_fActive = false;
			}
//			else if (wParam_ == SC_MOVE)
//				g_fActive = GetForegroundWindow() == hwnd_ && !IsIconic(g_hwnd);

			break;

		case WM_ENTERMENULOOP:
		case WM_EXITMENULOOP:
			fInMenu = (uMsg_ == WM_ENTERMENULOOP);
			return 0;

		// Application being activated or deactivated
		case WM_ACTIVATE:
//			g_fActive = wParam_ && !IsIconic(hwnd_);
			g_fActive = (wParam_ == WA_ACTIVE || wParam_ == WA_CLICKACTIVE);
			TRACE("WM_ACTIVATE[APP] (%d): now %s\n", wParam_, g_fActive ? "active" : "inactive");
			if (!g_fActive)
				Sound::Silence();
/*
				Sound::Stop();
			else
				Sound::Play();
*/
			Input::UpdateAcquire(g_fActive);
//			Video::UpdateOverlay();
			break;

//		case WM_KILLFOCUS:
			g_fActive = false;
			return 0;

		// Window is being sized or moved
		case WM_ENTERSIZEMOVE:
		case WM_EXITSIZEMOVE:
			// Keep quiet during sizing and moving
			if (uMsg_ == WM_ENTERSIZEMOVE)
				Sound::Silence();

			// Deactivate and release mouse input
			Input::UpdateAcquire(g_fActive = (uMsg_ == WM_EXITSIZEMOVE));
			break;


		// Menu is about to be activated
		case WM_INITMENU:
			UpdateMenuFromOptions();

			// Release the mouse so it can be used for moving the window if necessary
			Sound::Silence();
			Input::UpdateAcquire(false);
			break;

		case WM_PAINT:
		{
			if (GetUpdateRect(g_hwnd, NULL, FALSE))
			{
				PAINTSTRUCT ps;
				BeginPaint(hwnd_, &ps);

				Video::UpdateOverlay();

				if (!GetOption(fullscreen))
					Frame::Redraw();

				EndPaint(hwnd_, &ps);
				return 0;
			}
			break;
		}

		case WM_DISPLAYCHANGE:
		{
			TRACE("Display mode changed to %uBPP (%ux%u)\n", wParam_, LOWORD(lParam_), HIWORD(lParam_));

//			if (GetForegroundWindow() != hwnd_)
//				g_fActive = false;

			// Re-initialise the video unless the change was caused by us returning from fullscreen mode on exit
			if (!fClosing)
				Video::Init();
			break;
		}

		// System palette has changed
		case WM_PALETTECHANGED:
			// Don't react to our own palette changes!
			if (reinterpret_cast<HWND>(wParam_) == hwnd_)
				break;
			// fall through to WM_QUERYNEWPALETTE ...

		// We're about to get the focus which may require us to update our palette
		case WM_QUERYNEWPALETTE:
			// Realise our palette if necessary
			Video::UpdatePalette();
			return TRUE;

		case WM_MOVE:
		case WM_SIZE:
		{
			Video::UpdateOverlay();
			return 0;
		}

		case WM_SETCURSOR:
		{
			// Hide the cursor in full screen mode, unless the menu is active
			if (GetOption(fullscreen) && !fInMenu)
			{
				SetCursor(NULL);
				return TRUE;
			}
			break;
		}

		case WM_MOUSEMOVE:
			// Grab the mouse back as it's moving over the main window
			if (g_fActive != (g_fActive = (GetActiveWindow()==hwnd_)))
				Input::UpdateAcquire(g_fActive);
			break;

		case WM_KEYUP:
		case WM_KEYDOWN:
		{
			// Ignore key repeats, and key-ups for anything other than F12 (reset)
			if ((uMsg_ == WM_KEYDOWN && lParam_ & 0x40000000) || (uMsg_ == WM_KEYUP && ((wParam_ != VK_F12) && (wParam_ != VK_SUBTRACT))))
				break;

			// Process the emulator function keys
			switch (wParam_)
			{
/*
				case VK_F1:
					if (GetOption(drive1))
					{
						if (GetAsyncKeyState(VK_SHIFT) & 0x8000)
							g_Drive1.Eject();
						else
							InsertDisk(g_Drive1);
					}
					break;

				// F2 = insert disk 2, SHIFT-F2 = eject disk 2
				case VK_F2:
					if (GetOption(drive2) == 1)
					{
						if (GetAsyncKeyState(VK_SHIFT) & 0x8000)
							g_Drive2.Eject();
						else
							InsertDisk(g_Drive2);
					}
					break;


				// F3 = import data, SHIFT-F3 = export data
				case VK_F3:
					if (GetAsyncKeyState(VK_SHIFT) & 0x8000)
						DialogBoxParam(__hinstance, MAKEINTRESOURCE(IDD_EXPORT), g_hwnd, ImportExportDlgProc, NULL);
					else
						DialogBoxParam(__hinstance, MAKEINTRESOURCE(IDD_IMPORT), g_hwnd, ImportExportDlgProc, NULL);
					break;
*/

				// F12 = Z80 reset (can be held to view reset screens)
				case VK_F12:
				case VK_SUBTRACT:
					if (uMsg_ == WM_KEYDOWN)
					{
						CPU::Init(false);
						InterruptType = Z80_pause;
						Sound::Stop();
					}
					else
					{
						CPU::Init();
						InterruptType = Z80_none;
						Sound::Play();
					}
					break;

/*
				case VK_F2:
				{
					// This is the interrupt timing code posted to sam-users, which does similar
					// things to Defender in testing timing accuracy
					BYTE abLoop[] = {	0xf3, 0x3e, 0x80, 0xd3, 0xfe, 0x06, 0x08, 0x21, 0x00, 0x00,
										0xdb, 0xf9, 0xa0, 0xca, 0x0a, 0x80, 0xdb, 0xf9, 0xa0, 0xc2,
										0x10, 0x80, 0xdb, 0xf9, 0xa0, 0xca, 0x16, 0x80, 0x23, 0x00,
										0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
										0xdb, 0xf9, 0xa0, 0xc2, 0x1c, 0x80, 0xaf, 0xd3, 0xfe, 0xfb,
										0xe5, 0xc1, 0xc9 };

					memcpy(apbPageWritePtrs[1], abLoop, sizeof abLoop);
					break;
				}
*/

				case VK_F3:		SendMessage(g_hwnd, WM_COMMAND, IDM_HARDWARE_MOUSE_NONE + (GetOption(mouse)+1) % 3, 0L); break;

//				case VK_F5:		SendMessage(g_hwnd, WM_COMMAND, IDM_VIDEO_ACCURATE_MODE3, 0L); break;
				case VK_F6:		SendMessage(g_hwnd, WM_COMMAND, IDM_VIDEO_SCALING_1X1 + (GetOption(scale) % 3), 0L); break;
				case VK_F7:		SendMessage(g_hwnd, WM_COMMAND, IDM_VIDEO_SYNC_NONE + !GetOption(sync), 0L); break;
//				case VK_F8:		SendMessage(g_hwnd, WM_COMMAND, IDM_VIDEO_FULLSCREEN, 0L); break;

				case VK_F9:		SetOption(profile, !GetOption(profile)); break;

//				case VK_F11:	InterruptType = Z80_nmi; break;

//				case VK_PAUSE:	SendMessage(g_hwnd, WM_COMMAND, IDM_CPU_PAUSE, 0L); break;
			}
		}
		break;

		// Menu and commands (possible via the accelerator table)
		case WM_COMMAND:
		{
			WORD wId = LOWORD(wParam_);

			switch (wId)
			{
				// F1 = floppy 1 insert, Shift-F1 = floppy 1 eject
				case IDM_FILE_FLOPPY1_DEVICE:		g_Drive1.Insert("A:");	break;
				case IDM_FILE_FLOPPY1_INSERT:		InsertDisk(g_Drive1);	break;
				case IDM_FILE_FLOPPY1_EJECT:		g_Drive1.Eject();		break;
				case IDM_FILE_FLOPPY1_SAVE_CHANGES:	g_Drive1.Save();		break;

				// F2 = floppy 2 insert, Shift-F2 = floppy 2 eject
				case IDM_FILE_FLOPPY2_DEVICE:		g_Drive2.Insert("B:");	break;
				case IDM_FILE_FLOPPY2_INSERT:		InsertDisk(g_Drive2);	break;
				case IDM_FILE_FLOPPY2_EJECT:		g_Drive2.Eject();		break;
				case IDM_FILE_FLOPPY2_SAVE_CHANGES:	g_Drive2.Save();		break;

				// F4 = import data, Shift-F4 = export data
				case IDM_FILE_IMPORT_DATA:	DialogBoxParam(__hinstance, MAKEINTRESOURCE(IDD_IMPORT), g_hwnd, ImportExportDlgProc, NULL); break;
				case IDM_FILE_EXPORT_DATA:	DialogBoxParam(__hinstance, MAKEINTRESOURCE(IDD_EXPORT), g_hwnd, ImportExportDlgProc, NULL); break;

				// Alt-F4 = exit
				case IDM_FILE_EXIT:		PostMessage(hwnd_, WM_CLOSE, 0, 0L); break;

				// Items from help menu
				case IDM_HELP_GENERAL:	ShellExecute(hwnd_, NULL, "SimCoupe.txt", NULL, "", SW_SHOWMAXIMIZED); break;
				case IDM_HELP_FAQ:		ShellExecute(hwnd_, NULL, "FAQ.txt", NULL, "", SW_SHOWMAXIMIZED); break;
				case IDM_HELP_ABOUT:	DialogBoxParam(__hinstance, MAKEINTRESOURCE(IDD_ABOUT), g_hwnd, AboutDlgProc, NULL); break;

				// Most of the menu options are dealt with here
				default:	UpdateOptionsFromMenu(wParam_); break;
			}
			break;
		}
	}

	return DefWindowProc(hwnd_, uMsg_, wParam_, lParam_);
}


BOOL CALLBACK ImportExportDlgProc (HWND hdlg_, UINT uMsg_, WPARAM wParam_, LPARAM lParam_)
{
	static DWORD dwAddress = 32768, dwLength = 0;
	static BYTE bPage = 1;
	static WORD wOffset = 0;
	static char szFile [_MAX_PATH];
	static bool fBasicAddress = true, fUpdating = false;
	char sz[32];

	switch (uMsg_)
	{
		case WM_INITDIALOG:
		{
			CentreWindow(hdlg_);

			SetDlgItemText(hdlg_, IDE_FILE, szFile);
			wsprintf(sz, "%lu", dwAddress); SetDlgItemText(hdlg_, IDE_ADDRESS, sz);
//			wsprintf(sz, "%u", bPage);		SetDlgItemText(hdlg_, IDE_PAGE, sz);
//			wsprintf(sz, "%u", wOffset);	SetDlgItemText(hdlg_, IDE_OFFSET, sz);

			SendMessage(hdlg_, WM_COMMAND, fBasicAddress ? IDR_BASIC_ADDRESS : IDR_PAGE_OFFSET, 0); 
			return TRUE;
		}

		case WM_COMMAND:
		{
			WORD wControl = LOWORD(wParam_);
			bool fChange = HIWORD(wParam_) == EN_CHANGE;

			switch (wControl)
			{
				case IDCLOSE:
				case IDCANCEL:
					EndDialog(hdlg_, 0);
					return TRUE;

				case IDB_BROWSE:
				{
					bool fReadOnly = false;
					GetLoadFile(g_hwnd, "Binary files (*.bin)\0*.bin\0All files (*.*)\0*.*\0", NULL, szFile, sizeof szFile, &fReadOnly);
					SetDlgItemText(hdlg_, IDE_FILE, szFile);
					break;
				}

				case IDE_FILE:
					if (fChange)
					{
						GetDlgItemText(hdlg_, IDE_FILE, szFile, sizeof szFile);
						EnableWindow(GetDlgItem(hdlg_, IDOK), !(GetFileAttributes(szFile) & FILE_ATTRIBUTE_DIRECTORY));
					}
					break;

				case IDE_ADDRESS:
					if (fChange && !fUpdating)
					{
						GetDlgItemText(hdlg_, IDE_ADDRESS, sz, sizeof sz);
						dwAddress = strtoul(sz, NULL, 0) & 0x7ffff;
						bPage = static_cast<BYTE>(dwAddress/16384 - 1) & 0x1f;
						wOffset = static_cast<WORD>(dwAddress) & 0x3fff;

						fUpdating = true;
						wsprintf(sz, "%u", bPage);		SetDlgItemText(hdlg_, IDE_PAGE, sz);
						wsprintf(sz, "%u", wOffset);	SetDlgItemText(hdlg_, IDE_OFFSET, sz);
						fUpdating = false;
					}
					break;

				case IDE_PAGE:
				case IDE_OFFSET:
					if (fChange && !fUpdating)
					{
						GetDlgItemText(hdlg_, IDE_PAGE, sz, sizeof sz);
						bPage = static_cast<BYTE>(strtoul(sz, NULL, 0)) & 0x1f;

						GetDlgItemText(hdlg_, IDE_OFFSET, sz, sizeof sz);
						wOffset = static_cast<WORD>(strtoul(sz, NULL, 0)) & 0x3fff;

						fUpdating = true;
						wsprintf(sz, "%lu", static_cast<DWORD>(bPage + 1) * 16384 + wOffset);
						SetDlgItemText(hdlg_, IDE_ADDRESS, sz);
						fUpdating = false;
					}
					break;

				case IDR_BASIC_ADDRESS:
				case IDR_PAGE_OFFSET:
					fBasicAddress = wControl == IDR_BASIC_ADDRESS;
					SendDlgItemMessage(hdlg_, fBasicAddress ? IDR_BASIC_ADDRESS : IDR_PAGE_OFFSET, BM_SETCHECK, BST_CHECKED, 0); 

					EnableWindow(GetDlgItem(hdlg_, IDE_ADDRESS), fBasicAddress);
					EnableWindow(GetDlgItem(hdlg_, IDE_PAGE), !fBasicAddress);
					EnableWindow(GetDlgItem(hdlg_, IDE_OFFSET), !fBasicAddress);
					break;

				case IDOK:
				{
					GetDlgItemText(hdlg_, IDE_LENGTH, sz, sizeof sz);
					dwLength = strtoul(sz, NULL, 0) & 0x7ffff;

					FILE* hFile;
					if (!szFile[0] || !(hFile = fopen(szFile, "rb")))
						Message(msgError, "Invalid filename!");
					else
					{
						BYTE byte;
						while (fread(&byte, sizeof byte, 1, hFile))
						{
							apbPageWritePtrs[bPage][wOffset] = byte;
							if (!(++wOffset &= 0x3fff))
								++bPage &= 0x1f;
						}

						fclose(hFile);
						EndDialog(hdlg_, 1);
					}

					return TRUE;
				}
			}
		}
	}

	return FALSE;
}


bool InitWindow ()
{
	// set up and register window class
	WNDCLASS wc;
	ZeroMemory(&wc, sizeof wc);
	wc.style = CS_BYTEALIGNCLIENT | CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc = WindowProc;
	wc.hInstance = __hinstance;
	wc.hIcon = LoadIcon(__hinstance, MAKEINTRESOURCE(IDI_MAIN));
	wc.hbrBackground = reinterpret_cast<HBRUSH>(GetStockObject(BLACK_BRUSH));
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.lpszMenuName = MAKEINTRESOURCE(IDR_MENU);
	wc.lpszClassName = "SimCoupeClass";

	char szCaption[128];
	wsprintf(szCaption, "SimCoupe/Win32");// %d.%d%d", SIMCOUPE_Major, SIMCOUPE_Minor, SIMCOUPE_Patch);

	bool f = (RegisterClass(&wc) && (g_hwnd = CreateWindowEx(WS_EX_APPWINDOW, "SimCoupeClass", szCaption,
		WS_OVERLAPPEDWINDOW, CW_USEDEFAULT, CW_USEDEFAULT, 0, 0, NULL, NULL, __hinstance, NULL)));

	// Size the window for the mode specified in the options, then centre it
	ResizeWindow();
	CentreWindow(g_hwnd);

	// Load the keyboard accelerators
	g_hAccel = LoadAccelerators(__hinstance, MAKEINTRESOURCE(IDR_MAIN));


	// Create the 2 joystick menus with just a None option for now - will be filled in later!
	// These should be automatically freed when the main app menu is destroyed, since they're attached to it
	hmenuJoysticks1 = CreateMenu(), hmenuJoysticks2 = CreateMenu();
	AppendMenu(hmenuJoysticks1, MF_STRING, IDM_HARDWARE_JOYSTICK1_NONE, "&None");
	AppendMenu(hmenuJoysticks2, MF_STRING, IDM_HARDWARE_JOYSTICK2_NONE, "&None");
	ModifyMenu(GetMenu(g_hwnd), IDM_HARDWARE_JOYSTICK1, MF_BYCOMMAND|MF_POPUP|MF_STRING,
		reinterpret_cast<UINT>(hmenuJoysticks1), "&Joystick 1");
	ModifyMenu(GetMenu(g_hwnd), IDM_HARDWARE_JOYSTICK2, MF_BYCOMMAND|MF_POPUP|MF_STRING,
		reinterpret_cast<UINT>(hmenuJoysticks2), "Joystick 2");

	return f;
}
